package Jcg.graph.arraybased;

import java.util.*;

/**
 * Array based implementation of an Adjacency List Representation of a graph.
 * The graph is simple (no multiple edges), and there are not weights.
 * 
 * @author Luca Castelli Aleardi (Ecole Polytechnique, INF562)
 * 
 * @version feb 2014
 */
public class ArrayBasedAdjacencyListGraph implements ArrayBasedGraph {
	private int[][] neighbors; // for each vertex, stores the set of neighbors (their indices)
	private int[] degrees; // stores the degree of each vertex
	private static int maxDegree=100; // maximum vertex degree
	
	/**
	 * Create a graph with n vertices (with a given maximum vertex degree)
	 * 
	 * @param n the number of vertices
	 */
	public ArrayBasedAdjacencyListGraph(int n) {
		this.neighbors=new int[n][]; // create an array of n vertices
		this.degrees=new int[n]; // create an array for storing vertex degrees
		for(int i=0;i<n;i++)
			this.neighbors[i]=new int[maxDegree]; // initialize n adjacency lists for neighbors
	}
	
	/**
	 * Create a graph from the adjacency lists
	 * Each vertex stores the list of its neighbors (vertex indices)
	 * Remarks: 
	 * 	-) the graph is assumed to be undirected, and each edge appears twice: (u,v) and (v,u)
	 * 	-) vertex indices range from 0 to n-1
	 * 
	 * @param nodes
	 * 				an array containing (in the entry nodes[i]) the list of neighbors of vertex v_i
	 */
	public ArrayBasedAdjacencyListGraph(Collection<Integer>[] nodes) {
		int n=nodes.length; // number of nodes of the graph
		this.neighbors=new int[n][]; // create an array of n vertices
		this.degrees=new int[n]; // create an array for storing vertex degrees
		for(int i=0;i<n;i++) {
			this.degrees[i]=nodes[i].size();
			this.neighbors[i]=new int[nodes[i].size()]; // create the adjacency list for vertex v_i
			int count=0;
			for(Integer u: nodes[i]) { // iterate over the neighbors of vertex v_i
				this.neighbors[i][count]=u; // add the edge (v_i, u)
				count++;
			}
		}
	}

	/**
	 * Add an edge from vertex u to vertex v
	 * 
	 * @param u first extremity of an edge
	 * @param v second extremity of an edge
	 */
    public void addEdge(int u, int v) {
    	if(u<0 || v<0)
    		throw new Error("Error (addEdge method): wrong index (negative)");
    	if(u>=this.neighbors.length || v>= this.neighbors.length)
    		throw new Error("Error (addEdge method): wrong max degree (out of bounds)");
    	this.neighbors[u][this.degrees[u]]=v; // set v as neighbor of vertex u
    	this.degrees[u]++; // the degree of vertex u increases
    }
    
    public void removeEdge(int u, int v){
    	throw new Error("To be completed");
    }
    
	/**
	 * Check whether there is an edge from u to v
	 * 
	 * @param u first extremity of an edge
	 * @param v second extremity of an edge
	 * @return true whether there is an edge from u to v
	 */
    public boolean adjacent(int u, int v){
    	if(u<0 || u>=this.neighbors.length || v<0 || v>= this.neighbors.length)
    		throw new Error("Error (addEdge method): wrong vertex index");
    	for(int i=0;i<degree(u);i++) {
    		if(this.neighbors[u][i]==v)
    			return true;
    	}
    	return false;
    }
    
	/**
	 * Return the degree of a vertex
	 * 
	 * @param v the index of a vertex
	 * @return the degree of vertex v
	 */
    public int degree(int v){
    	if(v<0 || v>= this.neighbors.length)
    		throw new Error("Error (vertexDegree method): wrong vertex index");
    	return this.degrees[v];
    }
    
	/**
	 * Return the list of neighbors of a vertex
	 * 
	 * @param v the index of a vertex
	 * @return the list of neighbors of v
	 */
    public Collection<Integer> neighbors(int v) {
    	if(v<0 || v>= this.neighbors.length)
    		throw new Error("Error (vertexDegree method): wrong vertex index");
    	Collection<Integer> result=new LinkedList<Integer>();
    	for(int i=0;i<this.degrees[v];i++)
    		result.add(this.neighbors[v][i]);
    	return result;
    }
    
	/**
	 * Return the list of all (undirected) edges in the graph
	 * (each edge does appear only once)
	 * 
	 * @return the list of all edges
	 */
    public int[][] getEdges(){
    	int sumOfDegrees=0;
    	for(int i=0;i<this.degrees.length;i++)
    		sumOfDegrees=sumOfDegrees+this.degrees[i];
    	int nEdges=sumOfDegrees/2; // total number of edges
    	int[][] edges=new int[nEdges][2]; // the result (an array storing all edges)
    	int count=0;
    	for(int i=0;i<this.neighbors.length;i++) { // iterate over all vertices
    		for(int j=0;j<this.degrees[i];j++) { // iterate over the neighbors of vertex i
    			if(i<this.neighbors[i][j]) {
    				edges[count][0]=i;
    				edges[count][1]=this.neighbors[i][j];
    				count++;
    			}
    		}
    	}
    	return edges;
    }
    
    /**
	 * Give the number of vertices
	 * 
	 * @return the number of vertices
	 */
    public int sizeVertices(){
    	return this.neighbors.length;
    }
    
    /**
	 * Check whether the graph is combinatorially consistent
	 * 
	 */
    public boolean checkCombinatorics(){
    	System.out.print("Checking graph...");
    	if(this.degrees.length!=this.neighbors.length)
    		throw new Error("Error: wrong number of vertices");
    	
    	for(int i=0;i<this.neighbors.length;i++) {
    		if(this.neighbors[i].length==0) {
    			System.out.println("vertex "+i+" has empty adjacency list");
    			return false;
    		}
    	}
    	System.out.println("ok");
    	return true;
    }

    
}
