package Jcg.graph.arraybased;

import java.util.*;

/**
 * An implementation of a generic (undirected, possibly weighted) graph, using an adjacency matrix
 *
 * @author Luca Castelli Aleardi
 */
public class ArrayBasedAdjacencyMatrixGraph implements ArrayBasedGraph {

    int n=0; // number of vertices
    int[][] adjacency;
    double[][] weight;

    public ArrayBasedAdjacencyMatrixGraph() { }

    public ArrayBasedAdjacencyMatrixGraph(int n) {
    	this.n  = n;
    	this.adjacency=new int[n][n];
    	this.weight=new double[n][n];
    }

    public ArrayBasedAdjacencyMatrixGraph(int[][] adjacency) {
    	this.adjacency=adjacency;
    	this.n=adjacency[0].length;
    	this.weight=new double[n][n];
    }

    public ArrayBasedAdjacencyMatrixGraph(int[][] adjacency, double[][] weight) {
    	this.adjacency=adjacency;
    	this.weight=weight;
    	this.n=adjacency[0].length;
    }

    /**
     * add an edge between vertices d and a
     * By default, the edge weight is set to 1
     */
    public void addEdge(int d, int a) {
		if(d<0 || a<0) throw new Error("addEdge: vertex index error");
		adjacency[d][a]=1;
		adjacency[a][d]=1;
		weight[d][a]=1.;
		weight[a][d]=1.;
    }

    /**
     * add an edge with a given weight
     */
    public void addEdge(int d, int a, double weight) {
		if(d<0 || a<0) throw new Error("addEdge: vertex index error");
		adjacency[d][a]=1;
		adjacency[a][d]=1;
		this.weight[d][a]=weight;
		this.weight[a][d]=weight;
    }

    /**
     * add an edge with a given weight
     */
    public void setWeight(int d, int a, double weight) {
		if(d<0 || a<0) throw new Error("setWeight: vertex index error");
		this.weight[d][a]=weight;
		this.weight[a][d]=weight;
    }

    /**
     * remove an edge between vertices d and a
     */
    public void removeEdge(int d, int a) {
		if(d<0 || a<0) throw new Error("removeEdge: vertex index error");
		adjacency[d][a]=0;
		adjacency[a][d]=0;
		weight[d][a]=0.;
		weight[a][d]=0.;
    }
    
    /**
     * test whether two vertices are adjacent in the graph
     */
    public boolean adjacent(int d, int a) {
		if(d<0 || a<0) throw new Error("adjacent: vertex index error");
		if(adjacency[d][a]==1 && adjacency[a][d]==1)
			return true;
		else if(adjacency[d][a]==0 && adjacency[a][d]==0)
			return false;
		else throw new Error("adjacent vertices: error");
    }

    /**
     * return the weight of edge (d,a)
     */
    public double getWeight(int d, int a) {
		if(d<0 || a<0) throw new Error("getWeight: vertex index error");
		if(weight==null) throw new Error("weight non defined");
		return this.weight[d][a];
    }

    /**
     * return the number of vertices of the graph
     */
    public int sizeVertices() {
    	return adjacency[0].length;
    }

    /**
     * return the degree of a node
     */
    public int degree(int index) {
    	int d=0;
    	for(int i=0;i<this.n;i++)
    		if(i!=index && adjacency[i][index]!=0) d++;
    	return d;
    }

    /**
     * return the vertices adjacent to a given vertex
     */
    public Collection<Integer> neighbors(int index) {
    	LinkedList<Integer> result=new LinkedList<Integer>();
    	for(int i=0;i<this.n;i++) {
    		if(i!=index && adjacency[i][index]!=0) {
    			result.add(i);
    		}
    	}
    	return result;
    }
 
    /**
     * return the list of edges (pair of vertices)
     */
    public int[][] getEdges() {
    	int numberOfEdges=0;
    	for(int i=0;i<this.n;i++)
    		numberOfEdges=numberOfEdges+degree(i);
    	numberOfEdges=numberOfEdges/2;
    	
    	int[][] result=new int[numberOfEdges][2];
    	int compt=0;
    	for(int i=0;i<this.n;i++) {
    		for(int j=i+1;j<this.n;j++){ 
    			if(adjacency[i][j]!=0) {
    				result[compt][0]=i;
    				result[compt][1]=j;
    				compt++;
    			}
    		}
    	}
    	return result;
    }

    public String toString() {
   		String result="adjacency matrix\n";
   		for(int i=0;i<sizeVertices();i++) {
   			for(int j=0;j<sizeVertices();j++)
   				result=result+" "+this.adjacency[i][j];
   			result=result+"\n";
   		}
   		result=result+"weights matrix\n";
   		for(int i=0;i<sizeVertices();i++) {
   			for(int j=0;j<sizeVertices();j++)
   				result=result+" "+this.weight[i][j];
   			result=result+"\n";
   		}
   		return result;
    }
                
}

