package Jcg.graph.arraybased;

import java.util.Collection;
import java.util.LinkedList;

import Jcg.geometry.Point_;
import Jcg.mesh.MeshLoader;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;

/**
 * This class provides input/output methods for reading (writing) graph
 * structures to (from) files
 * 
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2014)
 *
 */
public class ArrayBasedGraphLoader {

    /**
     * Load a graph from a file storing an edge list
     * Remark: the header (first line) contains the number of nodes and arcs
     * 
     * @param filename the input file encoding the graph
     * @return the resulting graph (array based adjacency list representation)
     */    
    public static ArrayBasedGraph readGraphFromFile(String filename){
    	System.out.print("Loading graph from file: ");
    	System.out.println(filename);
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	Jcg.io.IO.readTextFile(filename);
    	String line=Jcg.io.IO.readLine();
    	String[] w=Jcg.io.IO.wordsFromString(line);
    	int sizeVertices=Integer.parseInt(w[0]);
    	int sizeEdges=Integer.parseInt(w[1]);
    	ArrayBasedGraph g; // the resulting graph
    	g=new ArrayBasedAdjacencyListGraph(sizeVertices); // adjacency list representation
    	
    	System.out.print("\tAdding edges...");
    	int i=0;
    	while(i<sizeEdges) {
    		line=Jcg.io.IO.readLine();
    		w = Jcg.io.IO.wordsFromString(line);
    		if(w!=null && w.length>0 && w[0].charAt(0)!='#') { // skip empty lines and comments
    			int i1=Integer.parseInt(w[0]);
    			int i2=Integer.parseInt(w[1]);
    			g.addEdge(i1, i2);
    			g.addEdge(i2, i1); // LCA: added for ArrayBasedAdjacencyListGraph
    			i++;
    		}
    	}
    	System.out.println("done ("+sizeEdges+" edges)");
    	Jcg.io.IO.readStandardInput();
    	
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("Graph loaded: array based representation created ");
    	System.out.println(" ("+duration+" seconds)");
    	
    	return g;
    }
    
    /**
     * Load a graph from a file storing an edge list
     * Remark: there is no header
     * 
     * @param filename the input file encoding the graph
     * @return the resulting graph (array based adjacency list representation)
     */    
    public static ArrayBasedGraph readEdgeListGraphFromFile(String filename){
    	System.out.print("Loading graph from file (stored as edge list): ");
    	System.out.println(filename);

    	int sizeVertices=getNumberOfNodes(filename); // number of nodes of the graph
    	ArrayBasedAdjacencyListGraph g; // the resulting graph
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	LinkedList<Integer>[] neighbors=new LinkedList[sizeVertices]; // adjacency lists
    	for(int i=0;i<sizeVertices;i++)
    		neighbors[i]=new LinkedList<Integer>(); // create empty adjacency lists
    	
    	Jcg.io.IO.readTextFile(filename); // start reading the file
    	String line;
    	String[] w;
    	
    	int i=0; // count edges
    	while(Jcg.io.IO.endInput()==false) {
    		line=Jcg.io.IO.readLine();
    		w = Jcg.io.IO.wordsFromString(line); // parse line
    		if(w!=null && w.length>0 && w[0].charAt(0)!='#') { // skip empty lines and comments
    			int u=Integer.parseInt(w[0]);
    			int v=Integer.parseInt(w[1]);
    			neighbors[u].add(v); // add edge (u, v)
    			neighbors[v].add(u); // add edge (v, u)
    			i++;
    		}
    	}
     	g=new ArrayBasedAdjacencyListGraph(neighbors); // adjacency list representation
     	
    	Jcg.io.IO.readStandardInput();
    	
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("Graph loaded: array based representation created: ");
     	System.out.print(i+" arcs");
    	System.out.println(" ("+duration+" seconds)");
       	g.checkCombinatorics();
  	
    	return g;
    }
    
    /**
     * Get the number of nodes of a graph (stored as an edge list)
     * Remark: we assume there is no header (no number of nodes stored)
     * 
     * @param filename the input file encoding the graph
     * @return the number of nodes (1+ the maximum node index)
     */    
    public static int getNumberOfNodes(String filename){
    	int sizeOfNodes=-1; // result to return
    	//System.out.print("Reading file: ");
    	//System.out.println(filename);
    	//long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	Jcg.io.IO.readTextFile(filename);
    	
    	//System.out.print("\tReading lines...");
    	while(Jcg.io.IO.endInput()==false) {
    		String line=Jcg.io.IO.readLine();
    		String[] w = Jcg.io.IO.wordsFromString(line);
    		if(w!=null && w.length>0 && w[0].charAt(0)!='#') { // skip empty lines and comments
    			int i1=Integer.parseInt(w[0]);
    			int i2=Integer.parseInt(w[1]);
    			sizeOfNodes=Math.max(sizeOfNodes, i1);
    			sizeOfNodes=Math.max(sizeOfNodes, i2);
    		}
    	}
    	//System.out.println("done ("+(sizeOfNodes+1)+" nodes)");
    	Jcg.io.IO.readStandardInput();
    	
    	//endTime=System.nanoTime();
    	//double duration=(double)(endTime-startTime)/1000000000.;
    	//System.out.println(" ("+duration+" seconds)");
    	
    	return sizeOfNodes+1; // nodes have indices between 0 and n-1
    }

    /**
     * Load a graph from a file (in Metis format)
     * Warning: to check whether edges should appear once or twice
     * Remarks: 
     * 	-) vertex indices range from 1 to n (instead of 0..n-1)
     * 	-) edges appear only once in the file
     * 
     * @param filename the input file encoding the graph (lists of neighbors)
     * @return the resulting graph (array based adjacency list representation)
     */    
    public static ArrayBasedGraph readMetisFile(String filename){
    	System.out.println("Creating an array based representation of a graph from file (Metis format): "+filename);
    	ArrayBasedAdjacencyListGraph g; // the resulting graph
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	Jcg.io.IO.readTextFile(filename);
    	String line=Jcg.io.IO.readLine();
    	String[] w=Jcg.io.IO.wordsFromString(line);
    	int sizeVertices=Integer.parseInt(w[0]); // read the number of vertices
    	int sizeEdges=Integer.parseInt(w[1]); // read the number of arcs
    	
    	System.out.print("\tReading neighbors...");
    	LinkedList<Integer>[] neighbors=new LinkedList[sizeVertices];
    	for(int i=0;i<sizeVertices;i++)
    		neighbors[i]=new LinkedList<Integer>();
    	
    	for(int i=0; i<sizeVertices; i++) { // read the neighbors of vertex v_i from file
    		line=Jcg.io.IO.readLine();
    		if(line!=null && line.length()>0) {
    			w = Jcg.io.IO.wordsFromString(line); // w stores all neighbors of vertex i
    			if(w!=null && w.length>0) {
    				for(int j=0;j<w.length;j++) {
    					int u=Integer.parseInt(w[j]);
    					if(u<0 || u>sizeVertices)
    						throw new Error("Error reading graph from file: wrong vertex index"+u);
    					neighbors[i].add(u-1); // add edge (v_i, u)
    					//neighbors[u-1].add(i); // add opposite edge (u, v_i)
    				}
    			}
    		}
    	}
    
    	g=new ArrayBasedAdjacencyListGraph(neighbors); // adjacency list representation
    	System.out.println("done "+sizeEdges+" edges");
    	Jcg.io.IO.readStandardInput();
    	
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("array based representation created ("+duration+" seconds)");
    	
    	g.checkCombinatorics();
    	return g;
    }

    /**
     * Encode (the connectivity of) a given graph into a text file
     * 
     * @param g the graph to encode
     * @param filename the output file
     */    
    public static void writeGraphToFile(ArrayBasedGraph g, String filename){
    	int n=g.sizeVertices(); // number of vertices
    	int e=g.getEdges().length; // number of edges

    	System.out.print("Encoding the graph into a file ...");
    	Jcg.io.IO.writeNewTextFile(filename);
    	Jcg.io.IO.println(n+" "+e);
    	    	
    	// writing edges (a couple of indices for each edge)
    	int[][] edges=g.getEdges();
    	for(int i=0;i<edges.length;i++) {
    		int v1=edges[i][0];
    		int v2=edges[i][1];
    		Jcg.io.IO.println(v1+" "+v2);
    	}
    	
    	Jcg.io.IO.writeStandardOutput();
    	System.out.println("done ("+filename+")");
    }

    /**
     * Construct the skeleton graph from a mesh and encode it into a file
     * Warning: it works only for triangle meshes
     * 
     * @param meshname the file storing the mesh
     * @param output the output file (encoding the graph)
     */    
    public static void writeMeshSkeletonToFile(String meshname, String output){
    	Polyhedron_3 mesh=MeshLoader.getSurfaceMesh(meshname); // the input mesh
    	ArrayBasedGraph g=constructFromPolyhedron(mesh); // the 1-skeleton of the mesh
    	writeGraphToFile(g, output); // encode the 1-skeleton graph into a text file
    }

    /**
     * Construct the graph corresponding to the 1-skeleton of a 3D surface.
     * Graph vertices are ordered according to the original vertex numbering (in the polyhedron)
     */    
    public static ArrayBasedGraph constructFromPolyhedron(Polyhedron_3<Point_> polyhedron){
    	int size=polyhedron.sizeOfVertices();
    	ArrayBasedGraph g=new ArrayBasedAdjacencyListGraph(size);
    	
    	int count=0;
    	for(Vertex<Point_> v: polyhedron.vertices) { // assign label to vertices
    		v.index=count;
    		count++;
    	}
    	   	
    	for(Halfedge<Point_> e: polyhedron.halfedges) { // add all edges
    		int i1=e.getOpposite().getVertex().index; // source vertex
    		int i2=e.getVertex().index; // target index
    		//if(i1<i2) // add only one half-edge (i1, i2), when i1<i2
    			g.addEdge(i1, i2);
    	}
    	
    	return g;
    }

}
