package Jcg.polyhedron;

import java.util.*;

import Jcg.geometry.*;


/**
 * Pointer-based implementation of the Half-Edge data structure for polyhedral meshes
 * (allowing to represent planar and surface orientable meshes). <br>
 * <br>
 * This data structure allows us to represent both planar maps (embedded in R^2) and 3D surface meshes (embedded in R^3).
 * The generic type 'X' allows us to chose between the two cases: 'X' may be set either as 'Point_2' or 'Point_3'
 *
 * @author Code by Luca Castelli Aleardi (Ecole Polytechnique, INF562/INF555, 2010-2019)
 *
 * @see Primitives for the Manipulation of an orientable surface
 */
public class Polyhedron_3<X extends Point_>{

    public ArrayList<Vertex<X>> vertices;
    public ArrayList<Face<X>> facets;
    public ArrayList<Halfedge<X>> halfedges;

    public Polyhedron_3() {
        vertices=new ArrayList<Vertex<X>>();
        facets=new ArrayList<Face<X>>();
        halfedges=new ArrayList<Halfedge<X>>();
    }
   
    public Polyhedron_3(int n, int e, int f) {
        vertices=new ArrayList<Vertex<X>>(n);
        facets=new ArrayList<Face<X>>(f);
        halfedges=new ArrayList<Halfedge<X>>(e);
    }
    
    class ColorDecorator extends Decorator<Vertex<X>,Integer> {
    }
    
    public void DecorateVertices() {
    	ColorDecorator color=new ColorDecorator();
    	color.setDecoration(vertices.get(0),0);
    }

    public int sizeOfVertices() {
    	return this.vertices.size();
    }
    
    public int sizeOfFacets() {
    	return this.facets.size();
    }
    public int sizeOfHalfedges() {
    	return this.halfedges.size();
    }
    
    /**
     * compute the degree of a vertex
     */
    public int vertexDegree(Vertex<X> v) {
    	int result=0;
    	Halfedge<X> e=v.getHalfedge();
    	
    	Halfedge<X> pEdge=e.getNext().getOpposite();
    	while(pEdge!=e) {
    		pEdge=pEdge.getNext().getOpposite();
    		result++;
    	}
    	
    	return result+1;
    }
    
/**
 * returns true if there are no border edges
 */
	public boolean isClosed() {
		for(Halfedge<X> h: this.halfedges) {
			if(h.face==null)
				return false;
		}
		return true;
	}

/**
 * returns true if all vertices have exactly two incident edges
 */	
	public boolean isPureBivalent() {
		throw new Error("To be completed");
	}

/**
 * returns true if all vertices have exactly three incident edges
 */	
	public boolean isPureTrivalent() {
		throw new Error("a completer");
	}

	/**
	 * true iff the every face is a triangle
	 */	
	public boolean isPureTriangle() {
		for(Face<X> f: this.facets) {
			Halfedge<X> h=f.halfedge;
			if(h.next.next.next!=h)
				return false;
		}
		return true;
	}
	
	/**
	 * true iff the every face is a quadrangle
	 */	
	public boolean isPureQuad() {
		for(Face<X> f: this.facets) {
			Halfedge<X> h=f.halfedge;
			if(h.next.next==h) // degree two face (h is multiple edge)
				return false;
			if(h.next.next.next.next!=h)
				return false;
		}
		return true;
	}

/**
 * true iff the connected component denoted by h is a triangle
 */	
	public boolean isTriangle(Halfedge<X> h) {
		throw new Error("To be completed");
	}
   
    /**
     * Return the genus of the mesh (given by Euler formula)
     * Warning: works only for meshes with one connected component
     */
    public int genus() {
        int n=sizeOfVertices();
        int e=sizeOfHalfedges();
        int f=sizeOfFacets();
        int b=this.numberOfBoundaries();
        	return -(n-e/2+f-2+b)/2;
    }

    /**
     * Compute and return the number of boundaries
     */
    public int numberOfBoundaries() {
		//System.out.print("Processing boundaries of halfedge representation ...");
    	// store boundary edges (corresponding to half-edges without opposite reference)
    	HashSet<Halfedge<X>> isTreated=new HashSet<Halfedge<X>>(); // says whether a boundary edges has already been treated
    	LinkedList<Halfedge<X>> boundaryEdges=new LinkedList<Halfedge<X>>(); // stores boundary edges
        int nInsertedHalfedges=0; // counts the number of inserted halfedges
        
        for(Halfedge<X> e: this.halfedges) { // store all boundary half-edges
        	if(e.face==null)
        		boundaryEdges.add(e);
        }
        
        int bCount=0; // counts the number of boundaries
        while (boundaryEdges.isEmpty()==false) { // set opposite references of all boundary edges
        	Halfedge<X> firstEdge=boundaryEdges.poll(); // get a boundary
        	//System.out.println("processing edge "+printEdge(firstEdge));
        	if(isTreated.contains(firstEdge)==false) { // the edge belongs to a new boundary
        		isTreated.add(firstEdge); // mark the half-edge as treated
        		Halfedge<X> e=firstEdge.getNext();
        		while(e!=firstEdge) { // edge e turns around the boundary face
        			isTreated.add(e); // mark the half-edge as treated
        			e=e.getNext(); // visit next boundary half-edge
        		}
        		bCount++;
        	}
        }
        return bCount;
    }

/**
 * returns true if the polyhedral surface is combinatorially consistent. 
 * If borders==true normalization of the border edges is checked too. 
 * This method checks that each facet is at least a triangle and that 
 * the two incident facets of a non-border edge are distinct. 
 */	    
    public boolean isValid(boolean borders) {
    	boolean valid=true;
        System.out.print("Checking Polyhedron...");
        int n=this.vertices.size();
        int e=this.halfedges.size();
        int f=this.facets.size();

        int i=0;
        for(Halfedge<X> pedge: this.halfedges) {
        	if(pedge.getOpposite()==null || pedge.opposite.opposite!=pedge) { 
        		System.out.print("error opposite: "+i); 
        		valid=false; 
        	}
        	if(pedge.getNext()==null || pedge.next.prev!=pedge) { 
        		System.out.println("error next_edge: "+i); 
        		valid=false; 
        	}
        	if(pedge.getPrev()==null || pedge.prev.next!=pedge) { 
        		System.out.println("error prev_edge: "+i); 
        		valid=false; 
        	}
        	if(pedge.getVertex()==null) { 
        		System.out.println("error vertex: "+i); 
        		valid=false;
        	}
        	if(pedge.opposite==null || pedge.opposite.opposite==null || pedge.opposite.opposite!=pedge) {
        		System.out.println("error opposite edge: "+i); 
        		valid=false;        		
        	}
        	if(pedge.face!=null && pedge.face==pedge.getOpposite().face) { 
        		System.out.print("warning multiple edge: ");
        		System.out.println(" halfedge "+i+" is incident twice to face "+pedge.face);
        		valid=false; 
        	}
        	i++;
        }            
        
        i=0;
        for(Face<X> pface: this.facets){
        	if(pface==null) { System.out.println("error face pointer"); valid=false; }
        	if(pface.halfedge==null) { System.out.println("error face.halfedge"); valid=false; }
        	if(pface.degree()<3) { System.out.println("error face degree"); return valid=false; }
        	if(pface.halfedge.face!=pface) { System.out.println("error face.halfedge"); valid=false; }
        	i++;
        }
        
        i=0;
        for(Vertex<X> pvertex: this.vertices){
        	if(pvertex==null) { System.out.println("error vertex pointer:"+i); valid=false; }
        	if(pvertex.halfedge==null) { System.out.println("error vertex.halfedge: "+i); valid=false; }
        	if(pvertex.getPoint()==null) { System.out.println("error vertex.point: "+i); valid=false; }
        	if(pvertex.halfedge.vertex!=pvertex) { System.out.println("error vertex.halfedge: "+i); valid=false; }
        	i++;
        }            

        if(valid==true)
        	System.out.println("ok"); 
        else
         	System.out.println("not valid"); 
        
        if(this.isClosed())
        	System.out.println("Closed mesh: no boundaries");
        else
        	System.out.println("Mesh with boundaries");
        if(this.isPureTriangle())
        	System.out.println("The mesh is pure triangle");
        else if(this.isPureQuad())
        	System.out.println("The mesh is pure quad");
        else
        	System.out.println("The mesh is polygonal");
        
        int b=this.numberOfBoundaries();
        System.out.print("n: "+n+"  e: "+e/2+"  f: "+f+"  b:"+b);
        int g=-(n-e/2+f-2+b)/2;
        System.out.println("  genus: "+g);
        
        return valid;
    }
    
    /**
     * Perform the flip of an edge e=(u, v) <br>
     * Remarks: <br>
     * -) an edge can be flipped only if the two incident faces (u, v, w) and (u, v, z) are both triangles <br>
     * -) boundary edges cannot be flipped <br>
     * -) 'w' and 'z' must be not adjacent: otherwise we create a multiple edge
     */
    public void flipEdge(Halfedge<X> e) {
    	if(e==null)
    		throw new Error("Error: null edge");
    	if(e.getFace()==null || e.getOpposite().getFace()==null) // it cannot flip boundary edges
    		return;
    	
    	Vertex<X> v=e.getVertex();
    	Vertex<X> u=e.getOpposite().getVertex();
    	Vertex<X> w=e.getNext().getVertex();
    	Vertex<X> z=e.getOpposite().getNext().getVertex();
    	Face<X> f1=e.getFace();
    	Face<X> f2=e.getOpposite().getFace();
    	
    	if(f1.isTriangle()==false || f2.isTriangle()==false) // flips only concern triangle faces
    		return;
    	// before flipping we must check that 'w' and 'z' are not adjacent, to preserve combinatorial validity
    	List<Halfedge<X>> neighbors=w.getOutgoingHalfedges();
    	for(Halfedge<X> neighbor: neighbors) {
    		if(neighbor.getVertex()==z)
    			return;
    	}
    	
    	Halfedge<X> eVW=e.getNext(); // edge (V, W) oriented from V to W
    	Halfedge<X> eWU=e.getPrev();
    	Halfedge<X> eZV=e.getOpposite().getPrev(); // edge (Z, V)
    	Halfedge<X> eUZ=e.getOpposite().getNext();
    	
    	
    	// update vertices
    	u.setEdge(eWU);
    	v.setEdge(eZV);
    	w.setEdge(eVW);
    	z.setEdge(eUZ);
    	
    	// update the destination vertices of 'e' and its opposite half-edge
    	e.setVertex(w);
    	e.getOpposite().setVertex(z);
    	
    	eUZ.setFace(f1);
    	eVW.setFace(f2);
    	f1.setEdge(e);
    	f2.setEdge(e.getOpposite());
    	
    	// update the three halfedges inside F1
    	e.setNext(eWU);
    	eWU.setPrev(e);
    	eWU.setNext(eUZ);
    	eUZ.setPrev(eWU);
    	eUZ.setNext(e);
    	e.setPrev(eUZ);
    	
    	e.getOpposite().setNext(eZV);
    	eZV.setPrev(e.getOpposite());
    	eZV.setNext(eVW);
    	eVW.setPrev(eZV);
    	eVW.setNext(e.getOpposite());
    	e.getOpposite().setPrev(eVW);
    	
    	System.out.println("Edge e"+e.index+" flipped");
    }
    
    /**
     * Split a face by inserting a new vertex <p>
     * Warning: it works only for triangular faces<p>
     * 
     * Remark: faces are assumed to be ccw oriented (for combinatorics)
     */
    public void createCenterVertex(Face<X> f, X point){
    	int degree=f.degree();
    	if(degree!=3)
    		throw new Error("Error: the current implementation only works for triangle faces");
    	
    	// retrieve the three edges of the face
    	Halfedge<X> e=f.getEdge();
    	Halfedge<X> e1=e.getNext();
    	Halfedge<X> e2=e1.getNext();
    	
    	// retrieve the three vertices of the face
    	Vertex<X> v=e.getVertex(); // destination vertex of the edge 'e'
    	Vertex<X> u=e2.getVertex(); // source vertex of the edge'e'
    	Vertex<X> w=e1.getVertex();
    	
    	// create a new vertex
    	Vertex<X> newVertex=new Vertex<X>(point);
    	newVertex.index=this.sizeOfVertices();
    	this.vertices.add(newVertex); // the new vertex is added at the end of the list of vertices
    	
    	// create two new faces
    	Face<X> f1=new Face<X>();
    	f1.index=this.sizeOfFacets();
    	this.facets.add(f1); // the new face is added at the end of the list of faces
    	Face<X> f2=new Face<X>();
    	f2.index=this.sizeOfFacets();
    	this.facets.add(f2); // the new face is added at the end of the list of faces

    	// create 3 new pairs of halfedges
    	Halfedge<X> eUOut=new Halfedge<X>(); // halfedge outgoing from u
    	Halfedge<X> eUIn=new Halfedge<X>(); // halfedge oriented toward u
    	Halfedge<X> eVOut=new Halfedge<X>(); // halfedge outgoing from v
    	Halfedge<X> eVIn=new Halfedge<X>(); // halfedge oriented toward v
    	Halfedge<X> eWOut=new Halfedge<X>(); // halfedge outgoing from w
    	Halfedge<X> eWIn=new Halfedge<X>(); // halfedge oriented toward w
    	// add the new 6 halfedges to the mesh (setting their indices)
    	eUOut.index=this.sizeOfHalfedges();
    	this.halfedges.add(eUOut);
    	eUIn.index=this.sizeOfHalfedges();
    	this.halfedges.add(eUIn);
    	eVOut.index=this.sizeOfHalfedges();
    	this.halfedges.add(eVOut);
    	eVIn.index=this.sizeOfHalfedges();
    	this.halfedges.add(eVIn);
    	eWOut.index=this.sizeOfHalfedges();
    	this.halfedges.add(eWOut);
    	eWIn.index=this.sizeOfHalfedges();
    	this.halfedges.add(eWIn);
    	
    	// set references between faces and incident halfedges
    	f.halfedge=e;
    	f1.halfedge=e1;
    	f2.halfedge=e2;
    	e.face=f;
    	e1.face=f1;
    	e2.face=f2;
    	// set the two edges incident to the face f
    	eUIn.face=f;
    	eVOut.face=f;
    	// set the two edges incident to the face f1
    	eVIn.face=f1;
    	eWOut.face=f1;
    	// set the two edges incident to the face f2
    	eWIn.face=f2;
    	eUOut.face=f2;
    	
    	// set references between faces and incident vertices
    	newVertex.halfedge=eUOut;
    	eUIn.vertex=u;
    	eUOut.vertex=newVertex;
    	// set the two edges incident to the face f1
    	eVIn.vertex=v;
    	eVOut.vertex=newVertex;
    	// set the two edges incident to the face f2
    	eWIn.vertex=w;
    	eWOut.vertex=newVertex;
    	
    	// set 'prev' and 'next' references between halfedges, for the face 'f'
    	e.next=eVOut;
    	eVOut.prev=e;
    	e.prev=eUIn;
    	eUIn.next=e;
    	eUIn.prev=eVOut;
    	eVOut.next=eUIn;

    	// set 'prev' and 'next' references between halfedges, for the face 'f1'
    	e1.next=eWOut;
    	eWOut.prev=e1;
    	e1.prev=eVIn;
    	eVIn.next=e1;
    	eVIn.prev=eWOut;
    	eWOut.next=eVIn;
    	
    	// set 'prev' and 'next' references between halfedges, for the face 'f2'
    	e2.next=eUOut;
    	eUOut.prev=e2;
    	e2.prev=eWIn;
    	eWIn.next=e2;
    	eWIn.prev=eUOut;
    	eUOut.next=eWIn;
    	
    	// set 'opposite' references between 3 new pairs of halfedges
    	eVOut.opposite=eVIn;
    	eVIn.opposite=eVOut;
    	eWOut.opposite=eWIn;
    	eWIn.opposite=eWOut;
    	eUOut.opposite=eUIn;
    	eUIn.opposite=eUOut;
    }

    /**
     * replace the star of a vertex v by a new face 
     */	       
    public void eraseCenterVertex(Vertex<X> v) {
    	throw new Error("To be completed");
    }

    /**
     * Erase the vertex 'v' pointed by 'e', by replacing the star of 'v' by a new face. <p>
     * This method is not computationally efficient, requiring O(n) time: it remove all elements (6 halfedges, 1 vertex, 2 faces)
     * from the corresponding collections ('ArrayList'). <p>
     * <p>
     * Warning: <p>
     * -) it only does work with vertices having degree 3 <p>
     * -) the faces incident to vertex 'v' are assumed to exist (no holes incident to 'v') <p>
     * -) faces are assumed to be ccw oriented (only for combinatorics)
     * 
     * @return the halfedge 'h' which precedes 'e' in ccw direction (in the new triangular face)
     */	       
    public Halfedge eraseCenterVertex(Halfedge<X> e) {
    	Vertex v=e.getVertex();
    	if(this.vertexDegree(v)!=3) {
    		throw new Error("Error: the current implementation only works for vertices of degree 3");
    	}
    	
    	// get all 6 half-edges incident to vertex 'v' (they must be removed)
    	Halfedge<X> e1=e;
    	Halfedge g1=e.getOpposite();
    	Halfedge e2=g1.getPrev();
    	Halfedge g2=e2.getOpposite();
    	Halfedge e3=g2.getPrev();
    	Halfedge g3=e3.getOpposite();
    	
    	// get the 3 faces incident to vertex 'v' (f1 must be updated, f2 and f3 must be removed)
    	Face f1=e1.getFace();
    	Face f2=e2.getFace();
    	Face f3=e3.getFace();
    	if(f1==null || f2==null || f3==null)
    		throw new Error("Error: one of the incident faces is null (hole in the mesh)");
    	
    	// get the three vertices neighbors of 'v' (they must be updated)
    	Vertex v1=g1.getVertex();
    	Vertex v2=g2.getVertex();
    	Vertex v3=g3.getVertex();
    	
    	// get the three halfedges defining the new triangular face (v1, v2, v3) which must be "updated"
    	Halfedge h1=e1.getPrev();
    	Halfedge h2=e2.getPrev();
    	Halfedge h3=e3.getPrev();
    	
    	// update the references involving halfedges h1, h2 and h3
    	h1.next=h2;
    	h2.prev=h1;
    	h2.next=h3;
    	h3.prev=h2;
    	h3.next=h1;
    	h1.prev=h3;
    	
    	// h1, h2 and h3 are all incident to the "new" face f1
    	h1.face=f1;
    	h2.face=f1;
    	h3.face=f1;
    	
    	// vertices v1, v2 and v3 must be pointed by h1, h2 and h3
    	v1.halfedge=h1;
    	v2.halfedge=h2;
    	v3.halfedge=h3;
    	
    	// the face f1 must be incident to the halfedge h1
    	f1.halfedge=h1;
    	
    	// remove all elements
    	this.facets.remove(f2);
    	this.facets.remove(f3);
    	this.halfedges.remove(e1);
    	this.halfedges.remove(e2);
    	this.halfedges.remove(e3);
    	this.halfedges.remove(g1);
    	this.halfedges.remove(g2);
    	this.halfedges.remove(g3);
    	this.vertices.remove(v);
    	
    	this.resetMeshIndices();
    	
    	return h1; // return the halfedge preceding 'e=e1' in ccw direction
    }

    public Halfedge<X> makeHole(Halfedge<X> h){
    	if(h==null) throw new Error("error making hole: h null");
    	if(h.face==null) {
    		System.out.println("makeHole: h.face null");
    		return null;
    	}
    	
    	this.facets.remove(h.getFace());

    	Halfedge<X> p=h.next;
    	h.face=null;
    	while(p!=h) {
    		p.face=null;
    		p=p.next;
    	}
    	return h;
    }

    public Halfedge<X> fillHole(Halfedge h){
    	if(h.face!=null) throw new Error("error filling hole: h not boundary edge");
    	
    	Face<X> newFace=new Face<X>();
    	this.facets.add(newFace);
    	newFace.setEdge(h);

    	Halfedge p=h.next;
    	h.face=newFace;
    	int cont=1;
    	while(p!=h) {
    		p.face=newFace;
    		p=p.next;
    	}
    	return h;
    }

/**
 * Creates a new triangle facet within the hole incident to 'h' by connecting 
 * the tip of 'h' with two new halfedges and a new vertex. 
 * Returns the halfedge of the new edge that is incident to the new facet and the new vertex. 
 */	        
    public Halfedge<X> addTriangleToBorder(Halfedge h, X point) {
    	if(h.face!=null) 
    		throw new Error("no border edge");
    	//System.out.println("adding triangle to "+h);
    	
    	Face<X> newFace=new Face<X>();
    	Vertex<X> newVertex=new Vertex<X>(point);
    	Halfedge<X> hPrev=new Halfedge<X>();
    	Halfedge<X> hNext=new Halfedge<X>();
    	Halfedge<X> hPrevOpp=new Halfedge<X>();
    	Halfedge<X> hNextOpp=new Halfedge<X>();
    	
    	// setting the new face
    	newFace.setEdge(h);
    	// setting hPrev (halfedge preceding h in the new face)
    	hPrev.setFace(newFace);
    	hPrev.setVertex(h.getOpposite().getVertex());
    	hPrev.setPrev(hNext);
    	hPrev.setNext(h);
    	hPrev.setOpposite(hPrevOpp);
    	// setting hNext (halfedge following h in the new face)
    	hNext.setFace(newFace);
    	hNext.setVertex(newVertex);
    	hNext.setPrev(h);
    	hNext.setNext(hPrev);
    	hNext.setOpposite(hNextOpp);
    	// setting hPrevOpp (new boundary halfedge)
    	hPrevOpp.setFace(null);
    	hPrevOpp.setVertex(newVertex);
    	hPrevOpp.setPrev(h.getPrev());
    	hPrevOpp.setNext(hNextOpp);
    	hPrevOpp.setOpposite(hPrev);
    	// setting hNextOpp (the other new boundary halfedge)
    	hNextOpp.setFace(null);
    	hNextOpp.setVertex(h.getVertex());
    	hNextOpp.setPrev(hPrevOpp);
    	hNextOpp.setNext(h.getNext());
    	hNextOpp.setOpposite(hNext);
    	// updating old boundary halfedge informations
    	h.setFace(newFace);
    	h.setPrev(hPrev);
    	h.setNext(hNext);
    	// setting newVertex
    	newVertex.setEdge(hPrev); // LCA: a controler si c'est hPrev ou hNext
    	
    	// adding new facet, vertex and the four halfedges
    	this.vertices.add(newVertex);
    	this.facets.add(newFace);
    	this.halfedges.add(hPrev);
    	this.halfedges.add(hNext);
    	this.halfedges.add(hPrevOpp);
    	this.halfedges.add(hNextOpp);
    	
    	return hNext;
    }
    
 /**
 * a triangle with border edges is added to the polyhedral surface. 
 * Returns a non-border halfedge of the triangle.
 */	        
 	public Halfedge<X> makeTriangle(X p1, X p2, X p3) {    	
    	Face<X> newFace=new Face<X>();
    	Vertex<X> newVertex1=new Vertex<X>(p1);
    	Vertex<X> newVertex2=new Vertex<X>(p2);
    	Vertex<X> newVertex3=new Vertex<X>(p3);
    	Halfedge<X> e1=new Halfedge<X>();
    	Halfedge<X> e2=new Halfedge<X>();
    	Halfedge<X> e3=new Halfedge<X>();
    	Halfedge<X> e1Opp=new Halfedge<X>();
    	Halfedge<X> e2Opp=new Halfedge<X>();
    	Halfedge<X> e3Opp=new Halfedge<X>();
    	
    	// setting the new face
    	newFace.setEdge(e1);
    	// setting the new vertices (LCA: a' controler)
    	newVertex1.setEdge(e2);
    	newVertex2.setEdge(e3);
    	newVertex3.setEdge(e1);
    	// setting e1
    	e1.setFace(newFace);
    	e1.setVertex(newVertex1);
    	e1.setPrev(e3);
    	e1.setNext(e2);
    	e1.setOpposite(e1Opp);
    	// setting e2
    	e2.setFace(newFace);
    	e2.setVertex(newVertex2);
    	e2.setPrev(e1);
    	e2.setNext(e3);
    	e2.setOpposite(e2Opp);
    	// setting e3
    	e3.setFace(newFace);
    	e3.setVertex(newVertex3);
    	e3.setPrev(e2);
    	e3.setNext(e1);
    	e3.setOpposite(e3Opp);
    	// setting e1Opp (boundary halfedge opposite to e1)
    	e1Opp.setFace(null);
    	e1Opp.setVertex(newVertex3);
    	e1Opp.setPrev(e2Opp);
    	e1Opp.setNext(e3Opp);
    	e1Opp.setOpposite(e1);
    	// setting e2Opp (boundary halfedge opposite to e2)
    	e2Opp.setFace(null);
    	e2Opp.setVertex(newVertex1);
    	e2Opp.setPrev(e3Opp);
    	e2Opp.setNext(e1Opp);
    	e2Opp.setOpposite(e2);
    	// setting e3Opp (boundary halfedge opposite to e3)
    	e3Opp.setFace(null);
    	e3Opp.setVertex(newVertex2);
    	e3Opp.setPrev(e1Opp);
    	e3Opp.setNext(e2Opp);
    	e3Opp.setOpposite(e3);
    	
    	this.facets.add(newFace);
    	this.vertices.add(newVertex1);
    	this.vertices.add(newVertex2);
    	this.vertices.add(newVertex3);
    	this.halfedges.add(e1); this.halfedges.add(e1Opp);
    	this.halfedges.add(e2); this.halfedges.add(e2Opp);
    	this.halfedges.add(e3); this.halfedges.add(e3Opp);
    	
    	return e1;
    }

/**
 * splits the facet incident to h and g into two facets with a new diagonal 
 * between the two vertices denoted by h and g respectively. 
 * Returns h->next() after the operation, i.e., the new diagonal. The new face is to the right 
 * of the new diagonal, the old face is to the left. 
 * The time is proportional to the distance from h to g around the facet. 
 */	    
    public Halfedge<X> splitFacet(Halfedge<X> h, Halfedge<X> g) {
    	if(h==null || g==null) {
    		throw new Error("splitFacet: null pointer");
    	}
    	if(h.face==null || g.face==null) {
    		System.out.println("splitFacet: boundary edges");
    		return null;
    	}
    	if(h.face!=g.face) {
    		System.out.println("splitFacet: different incident facets");
    		return null;
    	}
    	if(h==g || h.next==g || g.next==h) {
    		System.out.println("splitFace error: loops and multiple edges are not allowed");
    		return null;
    	}

    	if(h.getNext().getNext().getNext()==h) {
    		System.out.println("splitFace error: triangular face");
    		return null;
    	}

		Halfedge<X> newDiagonal=new Halfedge<X>();
		Halfedge<X> oppositeNewDiagonal=new Halfedge<X>();
		Face<X> newFace=new Face<X>();

		// set face incidence relations
		newFace.setEdge(g);
		h.getFace().setEdge(h);
		newDiagonal.setFace(h.getFace());
		
		// set incidence relations between edges for the diagonals
		g.getNext().setPrev(newDiagonal);
		newDiagonal.setNext(g.getNext());
		h.getNext().setPrev(oppositeNewDiagonal);
		oppositeNewDiagonal.setNext(h.getNext());
		
		newDiagonal.setOpposite(oppositeNewDiagonal);
		oppositeNewDiagonal.setOpposite(newDiagonal);
		
		newDiagonal.setPrev(h);
		h.setNext(newDiagonal);
		
		oppositeNewDiagonal.setPrev(g);
		g.setNext(oppositeNewDiagonal);
		
		// set vertex incidence relations
		newDiagonal.setVertex(g.getVertex());
		oppositeNewDiagonal.setVertex(h.getVertex());
		
		// set face incidence relations for the new face
		g.setFace(newFace);
		Halfedge<X> pEdge=g.getNext();
		while(pEdge!=g) {
			pEdge.setFace(newFace);
			pEdge=pEdge.getNext();
		}
		
		// store the new cells (face and half-edges)
		this.halfedges.add(newDiagonal);
		this.halfedges.add(oppositeNewDiagonal);
		this.facets.add(newFace);
		
		return newDiagonal;
    } 

	/**
	 * Perform the collapse of edge e=(u,v)
	 * Edge e is shared by triangles (u,v,w) and (v,u,z)
	 */
	public void edgeCollapse(Halfedge<Point_3> e) {
		//System.out.print("Performing edge collapse...");
		if(e==null)
			return;
		
		// retrieve the cells incident to edge e
		Face<Point_3> f1=e.getFace();
		Face<Point_3> f2=e.getOpposite().getFace();
		Vertex<Point_3> u=e.getOpposite().getVertex();
		Vertex<Point_3> v=e.getVertex();
		Vertex<Point_3> w=e.getNext().getVertex();
		Vertex<Point_3> z=e.getOpposite().getNext().getVertex();
		Halfedge<Point_3> eUW=e.prev.opposite;
		Halfedge<Point_3> eWV=e.next.opposite;
		Halfedge<Point_3> eZU=e.opposite.next.opposite;
		Halfedge<Point_3> eVZ=e.opposite.prev.opposite;
		
		// update references between neighboring cells
		eUW.opposite=eWV;
		eWV.opposite=eUW;
		eZU.opposite=eVZ;
		eVZ.opposite=eZU;
		u.setEdge(eZU);
		w.setEdge(eUW);
		z.setEdge(eVZ);
		
		Halfedge<Point_3> pEdge=eWV;
		while(pEdge!=eVZ.getOpposite()) {
			pEdge.setVertex(u);
			pEdge=pEdge.getNext().getOpposite();
		}
		
		Point_3 newPoint=new Point_3(u.getPoint()); // compute new point location
		u.setPoint(newPoint);
		
		// remove old cells: 2 faces, 1 vertex, 6 halfedges
		// slow implementation: the arraylist must be updated at each removal
		this.vertices.remove(v);
		this.facets.remove(f1);
		this.facets.remove(f2);
		this.halfedges.remove(e);
		this.halfedges.remove(e.next);
		this.halfedges.remove(e.prev);
		this.halfedges.remove(e.opposite);
		this.halfedges.remove(e.opposite.next);
		this.halfedges.remove(e.opposite.prev);
		//System.out.println("done");
	}

    /**
     * splits the edge inserting a new vertex, with coordinates given by point. 
     */	    
        public Halfedge<X> splitEdge(Halfedge<X> h, X point) {
        	if(h==null) {
        		throw new Error("splitFacet: null pointer");
        	}
        	if(h.face==null || h.opposite.face==null) {
        		System.out.println("splitFacet: boundary edge... not yet implemented");
        		return null;
        	}
   		
    		// create the new edges, faces and vertex to be added
        	Halfedge<X> hNewLeft=new Halfedge<X>();
    		Halfedge<X> hNewRight=new Halfedge<X>();
    		Vertex<X> newVertex=new Vertex<X>(point);
    		
    		// set the vertex incidence relations
    		newVertex.setEdge(hNewLeft);
    		hNewLeft.setVertex(newVertex);

    		Vertex<X> u=h.getOpposite().getVertex();
    		hNewRight.setVertex(u);
    		u.setEdge(hNewRight);
    		
    		h.opposite.setVertex(newVertex);
    		
    		// set incidence relations for the first new halfedge
    		hNewLeft.setFace(h.face);
    		hNewLeft.setPrev(h.prev);
    		hNewLeft.setNext(h);
    		hNewLeft.setOpposite(hNewRight);
    		// warning: the order in the updates does count here
    		h.prev.setNext(hNewLeft); 
    		h.setPrev(hNewLeft);
    		
    		// set incidence relations for the second new halfedge
    		hNewRight.setFace(h.opposite.face);
    		hNewRight.setPrev(h.opposite);
    		hNewRight.setNext(h.opposite.next);
    		hNewRight.setOpposite(hNewLeft);
    		// warning: the order in the updates does count here
    		h.opposite.next.setPrev(hNewRight);
    		h.opposite.setNext(hNewRight);
    		
    		// add new cells to the polyhedron
    		this.vertices.add(newVertex);
    		this.halfedges.add(hNewLeft);
    		this.halfedges.add(hNewRight);
    		
    		return hNewLeft;
        } 

    /**
     * Reset indices for all: vertices, faces and halfedges. <p>
     * Indices correspond to the order of storage in the collections (e.g., vertex v0 is the first in the ArrayList 'vertices')
     */	
    public void resetMeshIndices() {
    	int count=0;
    	for(Vertex v: this.vertices) {
    		if(v==null)
    			throw new Error("Error: null vertex reference in array 'vertices'");
    		v.index=count;
    		count++;
    	}

    	count=0;
    	for(Face f: this.facets) {
    		if(f==null)
    			throw new Error("Error: null face reference in array 'facets'");
    		f.index=count; // standard face color
    		count++;
    	}

    	count=0;
    	for(Halfedge e: this.halfedges) {
       		if(e==null)
    			throw new Error("Error: null halfedge reference in array 'halfedges'");
    		e.index=count;
    		count++;
    	}
    }

    /**
     * Return a string representing the list of vertices
     */
    public String verticesToString() {
        String result="List of vertices\n";
        Iterator it=this.vertices.iterator();
        int cont=0;
        while(it.hasNext()) {
        	Vertex<X> v=(Vertex<X>)it.next();
        	result=result+"v"+cont+" "+v.getPoint().toString()+"\n"; 
        	cont++;
        }  	
        return result;           
    }

    /**
     * Return a string representing the list of faces
     */
    public String facesToString() {
        String result="List of faces\n";
        Iterator it=this.facets.iterator();
        int cont=0;
        while(it.hasNext()) {
        	Face<X> f=(Face<X>)it.next();
        	result=result+"f"+cont+" ";
        	/*Halfedge<X> e=f.getEdge();
        	while(e.getNext()!=f.getEdge()) {
        		result=result+vertices.indexOf(e.getVertex())+" "; 
        		e=e.getNext();
        	}
        	result=result+vertices.indexOf(e.getVertex())+"\n";*/
        	result=result+f.toString()+"\n";
        	cont++;
        }  	
        return result;           
    }

    /**
     * Return a string representing the list of faces
     */
    public String edgesToString() {
        String result="List of edges\n";
        Iterator it=this.halfedges.iterator();
        int cont=0;
        while(it.hasNext()) {
        	Halfedge<X> e=(Halfedge<X>)it.next();
        	result=result+"e"+cont+" ";
        	//result=result+vertices.indexOf(e.getVertex())+"\n";
        	result=result+e.toString()+"\n";
        	cont++;
        }  	
        return result;           
    }
    
}