package Jcg.schnyderwoods;

import java.util.List;

import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Vertex;

/**
 * @author Luca Castelli Aleardi (2019, Ecole Polytechnique)
 * 
 * A class for evaluating and optimizing the balance of a Schnyder wood orientation (for planar triangulations)
 */
public class Balance {
	
	private PlanarTriSchnyderWood sw;
	private int[][] incoming; // for each vertex, store the number of incoming edges for color 0, 1, 2
	public int[] degree;
	
	public int verbosity=1;
	
	public Balance(PlanarTriSchnyderWood sw) {
		System.out.print("Balance optimizer initialization...");
		this.sw=sw;
		int n=sw.polyhedron.sizeOfVertices();
		
		this.incoming=new int[n][3];
		this.degree=new int[n];
		this.initialize();
		System.out.println("done");
	}

    /**
     * Initialize the array 'incoming': for each vertex, count the number of incoming edges of color 0, 1 and 2
     */
	public void initialize() {
		int index;
		for(Vertex v: this.sw.polyhedron.vertices) {
			this.updateVertex(v);
		}
	}

    /**
     * Compute and store, for a given vertex 'v', the vertex degree and the number of incoming edges in each color
     */
	public void updateVertex(Vertex v) {
		int index=v.index;
		
		incoming[v.index][0]=0;
		incoming[v.index][1]=0;
		incoming[v.index][2]=0;
		
		this.degree[index]=this.sw.polyhedron.vertexDegree(v);
		if(v!=this.sw.v0 && v!=this.sw.v1 && v!=this.sw.v2) {
			List<Halfedge> edges=v.getOutgoingHalfedges(); // it can be improved here
			int red=0, blue=0, black=0;
			for(Halfedge e: edges) {
				if(this.sw.isWellOriented[e.index]==false && this.sw.edgeColor[e.index]==0)
					incoming[v.index][0]++;
				else if(this.sw.isWellOriented[e.index]==false && this.sw.edgeColor[e.index]==1)
					incoming[v.index][1]++;
				else if(this.sw.isWellOriented[e.index]==false && this.sw.edgeColor[e.index]==2)
					incoming[v.index][2]++;
			}
		}
	}

    /**
     * @return  the defect for vertex 'v'
     */
    public int defect(Vertex v) {
    	int min=0, max=0;
    	if(v!=this.sw.v0 && v!=this.sw.v1 && v!=this.sw.v2) {
    		min=Math.min(Math.min(this.incoming[v.index][0], this.incoming[v.index][1]), this.incoming[v.index][2]);
    		max=Math.max(Math.max(this.incoming[v.index][0], this.incoming[v.index][1]), this.incoming[v.index][2]);
    		
    		if(this.degree[v.index]%3==0) // if the degree is 3, 6, 9, ...
    			return (max-min);
    		else
    			return (max-min)-1;
    	}
    	else
    		return 0; // return 0 for v0, v1 or v2    	
    }

    /**
     * @return  true if the vertex 'v' is balanced
     */
    public boolean isBalanced(Vertex v) {
    	if(v!=this.sw.v0 && v!=this.sw.v1 && v!=this.sw.v2) {
    		if(this.defect(v)==0)
    			return true;
    		else
    			return false;
    	}
    	else
    		return false; 
    }

    /**
     * Perform the reversal of an oriented cycle if this leads to increase the number of balanced vertices
     */
    public int improveBalance(Halfedge e) {
    	if(this.sw.edgeColor[e.index]!=0) // consider only edges of color '0' (red)
    		return 0;
    	
    	Vertex u1=e.getVertex();
    	Vertex u2=e.next.getVertex();
    	Vertex u3=e.next.next.getVertex();
    	// check whether the face contains one vertex on the root face
    	if(u1==this.sw.v0 || u1==this.sw.v1 || u1==this.sw.v2)
    		return 0;
    	if(u2==this.sw.v0 || u2==this.sw.v1 || u2==this.sw.v2)
    		return 0;
    	if(u3==this.sw.v0 || u3==this.sw.v1 || u3==this.sw.v2)
    		return 0;
    	
    	int balancedVertices=0; // number of balanced vertices (before reversal)
    	int balancedVerticesAfter=0; // number of balanced vertices (after reversal)
    	
    	//System.out.println("e"+e.index+"\t"+this.sw.countOrientedTriangles()+"\tbalanced="+this.sw.countBalancedVertices());
    	
    	if(this.isBalanced(u1))
    		balancedVertices++;
    	if(this.isBalanced(u2))
    		balancedVertices++;
    	if(this.isBalanced(u3))
    		balancedVertices++;
    	
    	if(balancedVertices==3) // all 3 vertices are already balanced
    		return 0; // nothing to do

    	if(this.sw.isCCWOriented(e)) {
    		//System.out.print("\tbalanced="+this.sw.countBalancedVertices());
    		/*this.sw.reverseCCWTriangle(e);
			this.updateVertex(u1);
			this.updateVertex(u2);
			this.updateVertex(u3);*/  		
    		//System.out.print("\t"+this.sw.countOrientedTriangles()+"\t"+this.sw.countBalancedVertices());
    		// simulate the reversal of the triangle face containing 'e'
    		this.incoming[u1.index][0]--;
    		this.incoming[u2.index][1]--;
    		this.incoming[u3.index][2]--;
    		this.incoming[u1.index][2]++;
    		this.incoming[u2.index][0]++;
    		this.incoming[u3.index][1]++;

    		if(this.isBalanced(u1))
    			balancedVerticesAfter++;
    		if(this.isBalanced(u2))
    			balancedVerticesAfter++;
    		if(this.isBalanced(u3))
    			balancedVerticesAfter++;

    		if(balancedVerticesAfter>balancedVertices) {// check whether the reversal improve the balance
    			this.sw.reverseCCWTriangle(e); // perform the reversal of the oriented triangle

    			this.updateVertex(u1);
    			this.updateVertex(u2);
    			this.updateVertex(u3);
    			//System.out.println("\timprovement");
    			return balancedVerticesAfter-balancedVertices; // improvement (number of new balanced vertices)
    		}
    		else {
        		//System.out.print("\t"+this.sw.countOrientedTriangles());
    			//this.sw.reverseCWTriangle(e);
    			this.updateVertex(u1);
    			this.updateVertex(u2);
    			this.updateVertex(u3);    		
    			//System.out.println("\tno improvement, balance="+this.sw.countBalancedVertices()+"\t"+this.sw.countOrientedTriangles());
    			return 0;
    		}
    	}
    	/*else if(this.sw.isCWOriented(e)) {
    		// simulate the reversal of the triangle face containing 'e'
    		this.incoming[u1.index][0]--;
    		this.incoming[u2.index][1]--;
    		this.incoming[u3.index][2]--;
    		this.incoming[u1.index][1]++;
    		this.incoming[u2.index][0]++;
    		this.incoming[u3.index][2]++;

    		if(this.isBalanced(u1))
    			balancedVerticesAfter++;
    		if(this.isBalanced(u2))
    			balancedVerticesAfter++;
    		if(this.isBalanced(u3))
    			balancedVerticesAfter++;

    		if(balancedVerticesAfter>balancedVertices) // check whether the reversal improve the balance
    			this.sw.reverseCWTriangle(e); // perform the reversal of the oriented triangle

    		this.updateVertex(u1);
    		this.updateVertex(u2);
    		this.updateVertex(u3);
    	}*/
    	return 0;
    }

    /**
     * Perform the reversal of an oriented cycle if this leads to increase the number of balanced vertices<br>
     * 
     * For debugging
     */
    public int improveBalanceSlow(Halfedge e) {
    	if(this.sw.edgeColor[e.index]!=0) // consider only edges of color '0' (red)
    		return 0;
    	
    	Vertex u1=e.getVertex();
    	Vertex u2=e.next.getVertex();
    	Vertex u3=e.next.next.getVertex();
    	// check whether the face contains one vertex on the root face
    	if(u1==this.sw.v0 || u1==this.sw.v1 || u1==this.sw.v2)
    		return 0;
    	if(u2==this.sw.v0 || u2==this.sw.v1 || u2==this.sw.v2)
    		return 0;
    	if(u3==this.sw.v0 || u3==this.sw.v1 || u3==this.sw.v2)
    		return 0;
    	
    	int balancedVertices=0; // number of balanced vertices (before reversal)
    	int balancedVerticesAfter=0; // number of balanced vertices (after reversal)
    	
    	//System.out.println("e"+e.index+"\t"+this.sw.countOrientedTriangles()+"\tbalanced="+this.sw.countBalancedVertices());
    	
    	if(this.sw.isBalanced(u1))
    		balancedVertices++;
    	if(this.sw.isBalanced(u2))
    		balancedVertices++;
    	if(this.sw.isBalanced(u3))
    		balancedVertices++;
    	
    	if(balancedVertices==3) // all 3 vertices are already balanced
    		return 0; // nothing to do

    	if(this.sw.isCCWOriented(e)) {
    		//System.out.print("\tbalanced="+this.sw.countBalancedVertices());
    		this.sw.reverseCCWTriangle(e);
    		//System.out.print("\t"+this.sw.countOrientedTriangles()+"\t"+this.sw.countBalancedVertices());
    		if(this.sw.isBalanced(u1))
    			balancedVerticesAfter++;
    		if(this.sw.isBalanced(u2))
    			balancedVerticesAfter++;
    		if(this.sw.isBalanced(u3))
    			balancedVerticesAfter++;

    		if(balancedVerticesAfter>balancedVertices) {// check whether the reversal improve the balance
    			//System.out.println("\timprovement");
    			return balancedVerticesAfter-balancedVertices; // improvement (number of new balanced vertices)
    		}
    		else {
        		//System.out.print("\t"+this.sw.countOrientedTriangles());
    			this.sw.reverseCWTriangle(e);
    			//System.out.println("\tno improvement, balance="+this.sw.countBalancedVertices()+"\t"+this.sw.countOrientedTriangles());
    			return 0;
    		}
    	}
    	else if(this.sw.isCWOriented(e)) {
    		//System.out.print("\tbalanced="+this.sw.countBalancedVertices());
    		this.sw.reverseCWTriangle(e);
    		//System.out.print("\t"+this.sw.countOrientedTriangles()+"\t"+this.sw.countBalancedVertices());
    		if(this.sw.isBalanced(u1))
    			balancedVerticesAfter++;
    		if(this.sw.isBalanced(u2))
    			balancedVerticesAfter++;
    		if(this.sw.isBalanced(u3))
    			balancedVerticesAfter++;

    		if(balancedVerticesAfter>balancedVertices) {// check whether the reversal improve the balance
    			//System.out.println("\timprovement");
    			return balancedVerticesAfter-balancedVertices; // improvement (number of new balanced vertices)
    		}
    		else {
        		//System.out.print("\t"+this.sw.countOrientedTriangles());
    			this.sw.reverseCCWTriangle(e);
    			//System.out.println("\tno improvement, balance="+this.sw.countBalancedVertices()+"\t"+this.sw.countOrientedTriangles());
    			return 0;
    		}
    	}
    	return 0;
    }

    /**
     * Perform the reversal of oiented cycles to increase the number of balanced vertices
     */
    public int improveBalanceOnePass() {
    	System.out.print("One pass optimization...");
    	
    	int improvement=0;
    	for(Halfedge e: this.sw.polyhedron.halfedges) {
    		//improvement=improvement+this.improveBalance(e);
    		improvement=improvement+this.improveBalanceSlow(e);
    	}
    	
    	System.out.println("done\t improvement="+improvement+", balanced vertices="+this.sw.countBalancedVertices());
    	return improvement;
    }

    /**
     * Perform the reversal of oriented cycles to increase the number of balanced vertices <br>
     * 
     * Remark: <br>
     * -) the reversal process is performed until the absolute percentage of improvement is below a given threshold
     */
    public void improveBalance(double threshold) {
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	System.out.println("Starting post-processing: balance optimization");
    	double n=this.sw.polyhedron.sizeOfVertices();
    	
    	int improvement=this.improveBalanceOnePass();
    	double proportion=improvement/n;
    	int totalImprovement=0;
    	while(proportion>threshold) {
    		totalImprovement=totalImprovement+improvement;
    		improvement=this.improveBalanceOnePass();
    		proportion=improvement/n;
    	}
    	
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        System.out.println("Balance optimization ...done ("+duration+" seconds)");
    }
    
    /**
     * @return a real value, evaluating the balance of the edge orientation
     */
    public double getAverageDefect() {
    	double n=this.sw.polyhedron.sizeOfVertices();
    	this.initialize();
    	
    	double balance=0.;
    	for(Vertex v: this.sw.polyhedron.vertices) {
    		if(v!=this.sw.v0 && v!=this.sw.v1 && v!=this.sw.v2) {
    			int vertexDefect=this.defect(v);
    			balance=balance+vertexDefect;
    		}
    	}
    	return balance/(n-3); // we do not count the 3 vertices on the root outer face
    }
    
    /**
     * @return a real value, evaluating the balance of the edge orientation
     */
    public double countAverageDefect() {
    	double n=this.sw.polyhedron.sizeOfVertices();
    	
    	double balance=0.;
    	
    	int countNotDefecting=0;
    	int countDefect1=0;
    	int countDefect2=0;
    	int countDefect3=0;
    	int vertexDefect;
    	for(Vertex v: this.sw.polyhedron.vertices) {
    		if(v!=this.sw.v0 && v!=this.sw.v1 && v!=this.sw.v2) {
    			vertexDefect=defect(v);
    			if(vertexDefect==0)
    				countNotDefecting++; // count vertices whose defect is not zero
    			if(vertexDefect==1)
    				countDefect1++; // count vertices whose defect is not zero
    			if(vertexDefect==2)
    				countDefect2++; // count vertices whose defect is not zero
    			if(vertexDefect==3)
    				countDefect3++; // count vertices whose defect is not zero
    			
    			balance=balance+vertexDefect;
    		}
    	}
    	System.out.println("Number of non-defecting vertices: "+countNotDefecting+", defect 1: "+countDefect1+", defect 2: "+countDefect2+", defect 3: "+countDefect3);
    	return balance/(n-3); // we do not count the 3 vertices on the root outer face
    }

}
