package Jcg.schnyderwoods;

import Jcg.geometry.Point_;
import Jcg.geometry.Point_2;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Vertex;
import arraybased.meshes.MinimalHalfedgeDS;

public class FastSchnyderDrawing {
	int verbosity=0;
	
	/** Array-based representation of a planar triangulation (half-edge interface) */
	MinimalHalfedgeDS he;
	/** Schnyder wood (edge) orientation */
	FastSchnyderWood sw=null;
	/** number of vertices */
	int n;
	
	/** store the 2d coordinates of the Schnyder drawing*/
	Point_2[] coord2D=null;
	
	/**
	 * A table of size 8*n storing all (vertex) data involved in the computation of the Schnyder drawing <br>
	 * -) the first 3 columns, 0-2, store the height of nodes in the trees T0, T1 and T2 <br>
	 * -) columns 3, 4 and 5 store the cumulative size of the sub-trees rooted along the paths to the root s<br>
	 * -) columns 6 and 7 store the size of the subtrees rooted on T0 and T2 (red and black trees)
	 * */
	public int[] data;
	
	/** store, for each inner vertex 'v', the cumulative size of sub-trees of color black (in T2) rooted on the path P1[v] */
	public int[] dataP1T2;
	/** store, for each inner vertex 'v', the cumulative size of sub-trees of color red (in T0) rooted on the path P2[v] */
	public int[] dataP2T0;
	
	public FastSchnyderDrawing(FastSchnyderWood sw, int verbosity) {
		this.verbosity=verbosity;
		this.sw=sw;
		this.he=this.sw.polyhedron;
        Runtime runtime = Runtime.getRuntime();
        runtime.gc();
        long usedMemory1 = runtime.totalMemory() - runtime.freeMemory();
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

		// initialize data for processing the Schnyder drawing
		this.n=sw.polyhedron.sizeOfVertices();
		this.data=new int[8*this.n];
		this.dataP1T2=new int[n];
		this.dataP2T0=new int[n];
		
		long usedMemory2 = runtime.totalMemory() - runtime.freeMemory();
		long memory=(usedMemory2-usedMemory1); // Bytes
		long memoryMB=(usedMemory2-usedMemory1)/(1024L*1024L); // MBytes
		if(verbosity>0)
			System.out.println("Schnyder drawing initialization...done (memory usage: "+memoryMB+" MBytes)");
	}
	
	public void init(FastSchnyderWood sw) {
		this.sw=sw;
		
		for(int i=0;i<8*this.n;i++)
			this.data[i]=0;
	}
	
	/**
	 * Return the total number of nodes in a given 'region', including boundary vertices (recall there are 3 disjoint regions)
	 * 
	 * @param v  the node index
	 * @param region  an index between 0, 1, 2
	 * @return  the number of nodes (inner nodes and boundary nodes) of a given 'region'
	 */
	public int getVertexArea(int v, int region) {
		if(v==this.sw.v0 || v== this.sw.v1 || v==this.sw.v2) // LCA: not necessary instruction (can be removed for benchmarks
			return 0;
		if(region==2)
			return this.getCumulativeSize(v, 2)+this.dataP1T2[v]+2-this.getSubTreeSize(v, 2);
		else if(region==0)
			return this.getCumulativeSize(v, 0)+this.dataP2T0[v]+2-this.getSubTreeSize(v, 0);
		else {
			int result=n-(this.getVertexArea(v, 0)+this.getVertexArea(v, 2)-(this.getNodeHeight(v, 1)+1));
			result=result+this.getBoundarySize(v, 1); // LCA: it can be improved here (for runtime performance)
			return result;
		}
	}

	/**
	 * Return the number of inner nodes in a given 'region' (recall there are 3 disjoint regions)
	 * 
	 * @param v  the node index
	 * @param region  an index between 0, 1, 2
	 * @return  the number of nodes (inner nodes and boundary nodes) of a given 'region'
	 */
	public int getInnerVertexArea(int v, int region) {
		if(v==this.sw.v0 || v== this.sw.v1 || v==this.sw.v2) // LCA: not necessary instruction (can be removed for benchmarks
			return 0;
		if(region==2)
			return (this.getCumulativeSize(v, 2)+this.dataP1T2[v]+2-this.getSubTreeSize(v, 2))-this.getBoundarySize(v, 2);
		else if(region==0)
			return (this.getCumulativeSize(v, 0)+this.dataP2T0[v]+2-this.getSubTreeSize(v, 0))-this.getBoundarySize(v, 0);
		else {
			int result=n-(this.getVertexArea(v, 0)+this.getVertexArea(v, 2)-(this.getNodeHeight(v, 1)+1));
			result=result+this.getBoundarySize(v, 1); // LCA: it can be improved here (for runtime performance)
			return result-this.getBoundarySize(v, 1);
		}
	}

	/**
	 * Return the number of boundary nodes of a given 'region' (recall there are 3 disjoint regions)
	 * 
	 * @param v  the node index
	 * @param region  an index between 0, 1, 2
	 * @return  the number of boundary nodes of a given 'region'
	 */
	public int getBoundarySize(int v, int region) {
			return this.getNodeHeight(v, (region+1)%3)+this.getNodeHeight(v, (region+2)%3)+1;
	}

	/**
	 * Return the total number of faces in a given 'region' (recall there are 3 disjoint regions)
	 * 
	 * @param v  the node index
	 * @param region  an index between 0, 1, 2
	 * @return  the number of faces inside a given 'region'
	 */
	public int getFaceArea(int v, int region) {
		return 2*this.getVertexArea(v, region)-(this.getBoundarySize(v, region)+2);
	}

	/**
	 * Store the size of the subtree of 'T_j' rooted at vertex 'v_i'
	 * 
	 * @param i  index of the vertex 'v_i'
	 * @param color  index of the tree T_j (j = 0, 1, 2)
	 * @param size  size of the subtree of a given tree T_j rooted at vertex 'v_i'
	 */
	public void setSubTreeSize(int i, int color, int size) {
			this.data[i*8+3+color]=size;
	}

	/**
	 * Return the size of the subtree of T_j rooted at vertex 'v_i'
	 * 
	 * @param i index of the vertex 'v_i'
	 * @param color  index of the tree T_j (j = 0, 1, 2)
	 */
	public int getSubTreeSize(int i,int color) {
			return this.data[i*8+3+color];
	}

	/**
	 * Store the height of a vertex 'v' in the tree 'T_j'
	 * 
	 * @param i  index of the vertex 'v_i'
	 * @param color  index of the tree T_j (j = 0, 1, 2)
	 * @param height  height of the subtree of a given tree T_j rooted at vertex 'v_i'
	 */
	public void setNodeHeight(int i, int color, int height) {
			this.data[i*8+color]=height;
	}

	/**
	 * Return the height of a given node 'v' in the tree T_j
	 * 
	 * @param i index of the vertex 'v_i'
	 * @param color  index of the tree T_j (j = 0, 1, 2)
	 */
	public int getNodeHeight(int i,int color) {
			return this.data[i*8+color];
	}
	
	/**
	 * Store the cumulative size of all sub-trees (of a given 'color') rooted at the vertices along a path to the root
	 * 
	 * Warning: sub-tree size are not stored for the tree T1
	 * 
	 * @param i  index of the vertex 'v_i'
	 * @param color  color of the sub-trees (0, 1, 2)
	 * @param size  size of the subtree of a given tree T_j rooted at vertex 'v_i'
	 */
	public void setCumulativeSize(int i, int color, int size) {
		if(color==0)
			this.data[i*8+6]=size;
		else if(color==2)
			this.data[i*8+7]=size;
		else
			throw new Error("Error: cumulative sub-tree sizes are not stored for the tree T1");
	}

	/**
	 * Return the cumulative size
	 * 
	 * Warning: sub-tree size are not stored for the tree T1
	 * 
	 * @param i index of the vertex 'v_i'
	 * @param color  color of the subtrees (0, 1, 2)
	 */
	public int getCumulativeSize(int i,int color) {
		if(color==0)
			return this.data[i*8+6];
		else if(color==2)
			return this.data[i*8+7];
		else
			throw new Error("Error: cumulative sub-tree sizes are not stored for the tree T1");
	}

	/**
	 * Perform a DFS visit of the tree T0 and, for each inner vertex 'v', compute <br>
	 * -) the size of the sub-tree of color 0 (red) rooted at 'v' <br>
	 * -) the size of the sub-tree of color 2 (black) rooted at 'v' <br>
	 * 
	 * It performs the traversal of the contour of the tree, in ccw order.
	 * The traversal starts from the half-edge (u, v0) and ends at (v0, v2).
	 * (u, v0) is the half-edge oriented toward v0 corresponding to
	 * the first incoming edge incident to v0 (thus of color 0)
	 */
	public void computeSubtreeSizeT0T2() {
		//System.out.print("Computing the size of subtrees in T0...");
		//long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		for(int i=0;i<n;i++) {
			this.setSubTreeSize(i, 0, 1); // all the sub-trees in T0 have size 1 at the beginning
			this.setSubTreeSize(i, 2, 1); // all the sub-trees in T2 have size 1 at the beginning
		}
		this.setSubTreeSize(this.sw.v1, 0, 0); // v1 does not belong to T0
		this.setSubTreeSize(this.sw.v2, 0, 0); // v2 does not belong to T0
		this.setSubTreeSize(this.sw.v0, 2, 0); // v0 does not belong to T2
		this.setSubTreeSize(this.sw.v1, 2, 0); // v1 does not belong to T2

		int firstEdge=this.he.getPrev(this.sw.rootEdge); // starting half-edge, of color 0, oriented toward v_0
		int edge20=this.he.getOpposite(this.he.getNext(this.he.getOpposite(this.sw.rootEdge))); // halfedge v20, oriented toward v_0
		int lastEdge=this.he.getOpposite(edge20);
		int pEdge=firstEdge; 
		
		int node, size, descendant, ancestor;
    	while(pEdge!=lastEdge) { // perform the traversal of the contour of T_0
    		if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==true) { // an edge of 'T' visited the first time
    			pEdge=this.he.getPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==true) { // ']', a black incoming edge
    			//incomingDegrees=incomingDegrees+']';
    			node=this.he.getTarget(pEdge); // destination node in T2
    			descendant=this.he.getTarget(this.he.getOpposite(pEdge)); // origin of the black edge in T2
    			size=this.getSubTreeSize(descendant, 2); // size of the subtree in T2 (rooted at the descendant)
    			
    			this.setSubTreeSize(node, 2, this.getSubTreeSize(node, 2)+size); // increase the size, adding the size of a subtree
    			
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==false) { // outgoing blue edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==false) { // black outgoing edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==true) { // a blue incoming edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==false) { // an edge of T_0 encountered the second time
    			//treeCode=treeCode+")";
    			node=this.he.getTarget(pEdge);
    			size=this.getSubTreeSize(node, 0);
    			ancestor=this.he.getTarget(this.he.getOpposite(pEdge)); // ancestor node in the tree T0
    			
    			this.setSubTreeSize(ancestor, 0, this.getSubTreeSize(ancestor, 0)+size); // increment the size of the subtree rooted at the ancestor node
    			
    			pEdge=this.he.getPrev(pEdge);
    		}
    	}
    	// we "must" set the subtree size at 'v2'
		this.setSubTreeSize(this.sw.v2, 2, n-2); // v2 has a sub-tree of size 'n-2'
    	
    	//endTime=System.nanoTime();
        //double duration=(double)(endTime-startTime)/1000000000.;
    	//System.out.print("done");
    	//System.out.println(" ("+duration+" seconds)");
	}

	/**
	 * Perform a DFS visit of the tree T0: for each inner vertex 'v' compute and store in a single pass <br>
	 * -) the height of node 'v' in the tree T_0 <br>
	 * -) the cumulative size of all subtrees in 'T_2' rooted at vertex 'v' (along the path from 'v' to 'v0') <br>
	 * 
	 * <br>
	 * It performs the traversal of the contour of the tree, in ccw order.
	 * The traversal starts from the half-edge (u, v0) and ends at (v0, v2).
	 * (u, v0) is the half-edge oriented toward v0 corresponding to
	 * the first incoming edge incident to v0 (thus of color 0)
	 */
	public void computeHeightT0() {
		//System.out.print("Computing the height of nodes in the tree T0...");
		//long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int firstEdge=this.he.getPrev(this.sw.rootEdge); // starting half-edge, of color 0, oriented toward v_0
		int edge20=this.he.getOpposite(this.he.getNext(this.he.getOpposite(this.sw.rootEdge))); // halfedge v20, oriented toward v_0
		int lastEdge=this.he.getOpposite(edge20);
		int pEdge=firstEdge; 
		
		int height=0;
    	while(pEdge!=lastEdge) { // perform the traversal of the contour of T_0
    		if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==true) { // an edge of 'T' visited the first time
    			if(pEdge!=edge20) {
    				height++;
    				
    				int newNode=this.he.getTarget(this.he.getOpposite(pEdge)); // index of the new node encountered during the DFS
    				int ancestor=this.he.getTarget(pEdge); // index of its ancestor node
    				
    				this.setNodeHeight(newNode, 0, height); // store the height of the new (descendant) node in 'T0'
    				
    				// compute the cumulative size of sub-trees of color '2' (black) rooted on the 'red' path (from the node to the root of 'T0')
    				int cumulative=this.getCumulativeSize(ancestor, 2);
    				this.setCumulativeSize(newNode, 2, cumulative+this.getSubTreeSize(newNode, 2));
    			}
    			if(pEdge!=firstEdge) { // '['
    			}
    			pEdge=this.he.getPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==true) { // a black incoming edge ']'
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==false) { // outgoing blue edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==false) { // black outgoing edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==true) { // a blue incoming edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==false) { // ')', an edge of T_0 encountered the second time
    			height--; 
    			
    			pEdge=this.he.getPrev(pEdge);
    		}
    	}
    	
    	//endTime=System.nanoTime();
        //double duration=(double)(endTime-startTime)/1000000000.;
    	//System.out.print("done");
    	//System.out.println(" ("+duration+" seconds)");
	}

	/**
	 * Perform a DFS visit of the tree T1: for each inner vertex 'v' compute and store in a single pass <br>
	 * -) the height of node 'v' in the tree T_1 <br>
	 * -) the cumulative size of all subtrees in 'T_2' rooted at vertex 'v' (along the path from 'v' to 'v0') <br>
	 * 
	 * <br>
	 * It performs the traversal of the contour of the tree, in ccw order.
	 * The traversal starts from the half-edge (v2, v1) and ends at (v0, v1).
	 * (v2, v1) is the half-edge oriented toward v1 corresponding to
	 * the first incoming inner edge incident to v1 (thus of color 1)
	 */
	public void computeHeightT1() {
		/*System.out.print("Computing the height of nodes in the tree T1...");
		long startTime=System.nanoTime(), endTime; // for evaluating time performances*/
		
		int firstEdge=this.he.getPrev(this.he.getOpposite(this.he.getPrev(this.he.getOpposite(this.sw.rootEdge)))); // starting half-edge: inner halfedge of color 1, oriented toward v_1
		int edge01=this.sw.rootEdge; // halfedge v01, oriented toward v_1
		int lastEdge=edge01;
		int pEdge=firstEdge; 
		
		//System.out.println("First incoming blue edge (toward at v1): "+firstEdge.getOpposite().getVertex().index+", "+firstEdge.getVertex().index);
		
		int height=0;
		//int count=0;
    	while(pEdge!=lastEdge) { // perform the traversal of the contour of T_1
    		//count++;
    		//System.out.println("pEdge "+count+": "+pEdge.getOpposite().getVertex().index+", "+pEdge.getVertex().index);

    		if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==true) { // an edge of 'T1' visited the first time
    			if(pEdge!=edge01) {
    				height++;
    				
    				int newNode=this.he.getTarget(this.he.getOpposite(pEdge)); // index of the new node encountered during the DFS
    				int ancestor=this.he.getTarget(pEdge); // index of its ancestor node
    				
    				this.setNodeHeight(newNode, 1, height); // store the height of the new (descendant) node in 'T1'
    				
    				// compute the cumulative size of sub-trees of color '0' rooted on the 'blue' path (from the node to the root of 'T1')
    				int cumulative=this.getCumulativeSize(ancestor, 0);
    				this.setCumulativeSize(newNode, 0, cumulative+this.getSubTreeSize(newNode, 0));

    				// compute the cumulative size of sub-trees of color '2' rooted on the 'blue' path (from the node to the root of 'T1')
    				//cumulative=this.getCumulativeSize(ancestor, 2);
    				//this.setCumulativeSize(newNode, 2, cumulative+this.getSubTreeSize(newNode, 2));
    				cumulative=this.dataP1T2[ancestor];
    				this.dataP1T2[newNode]=cumulative+this.getSubTreeSize(newNode, 2);
    			}
    			pEdge=this.he.getPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==true) { // a black incoming edge ']'
    			//incomingDegrees=incomingDegrees+']';
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==false) { // outgoing blue edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==false) { // black outgoing edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==true) { // a blue incoming edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==false) { // ')', an edge of T_0 encountered the second time
    			//treeCode=treeCode+")";
    			height--; 
    			
    			pEdge=this.he.getPrev(pEdge);
    		}
    	}
    	
    	/*endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("done");
    	System.out.println(" ("+duration+" seconds)");*/
	}

	/**
	 * Perform a DFS visit of the tree T2: for each inner vertex 'v' compute and store in a single pass <br>
	 * -) the height of node 'v' in the tree T_2 <br>
	 * -) the cumulative size of all subtrees in 'T_0' rooted at vertex 'v' (along the path from 'v' to 'v2') <br>
	 * 
	 * <br>
	 * It performs the traversal of the contour of the tree, in ccw order. <br>
	 * <br>
	 * Remark: 
	 * the first incoming inner edge incident to v2 (thus of color 2)
	 */
	public void computeHeightT2() {
		/*System.out.print("Computing the height of nodes in the tree T2...");
		long startTime=System.nanoTime(), endTime; // for evaluating time performances*/
		
		int firstEdge=this.he.getPrev(this.he.getOpposite(this.he.getNext(this.he.getOpposite(this.sw.rootEdge)))); // starting half-edge: inner halfedge of color 2, oriented toward v_2
		int edge12=this.he.getPrev(this.he.getPrev(this.he.getOpposite(this.sw.rootEdge))); // halfedge v12, oriented toward v_2
		int lastEdge=edge12;
		int pEdge=firstEdge; 
		
		//System.out.println("First incoming blue edge (toward at v1): "+firstEdge.getOpposite().getVertex().index+", "+firstEdge.getVertex().index);
		
		int height=0;
		int cumulative;
		//int count=0;
    	while(pEdge!=lastEdge) { // perform the traversal of the contour of T_1
    		//count++;
    		//System.out.println("pEdge "+count+": "+pEdge.getOpposite().getVertex().index+", "+pEdge.getVertex().index);

    		if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==true) { // an edge of 'T2' visited the first time
    			if(pEdge!=edge12) {
    				height++;
    				
    				int newNode=this.he.getTarget(this.he.getOpposite(pEdge)); // index of the new node encountered during the DFS
    				int ancestor=this.he.getTarget(pEdge); // index of its ancestor node
    				
    				this.setNodeHeight(newNode, 2, height); // store the height of the new (descendant) node in 'T2'
    				
    				// compute the cumulative size of sub-trees of color '0' rooted on the 'black' path (from the node to the root of 'T2')
    				//cumulative=this.getCumulativeSize(ancestor, 0);
    				//this.setCumulativeSize(newNode, 0, cumulative+this.getSubTreeSize(newNode, 0));
    				cumulative=this.dataP2T0[ancestor];
    				this.dataP2T0[newNode]=cumulative+this.getSubTreeSize(newNode, 0);
    			}
    			pEdge=this.he.getPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==true) { // a black incoming edge ']'
    			//incomingDegrees=incomingDegrees+']';
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==false) { // outgoing blue edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==false) { // black outgoing edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==true) { // a blue incoming edge
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==false) { // ')', an edge of T_0 encountered the second time
    			//treeCode=treeCode+")";
    			height--; 
    			
    			pEdge=this.he.getPrev(pEdge);
    		}
    	}
    	
    	/*endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("done");
    	System.out.println(" ("+duration+" seconds)");*/
	}

	/**
	 * Perform a DFS visit of the tree T1: for each inner vertex 'v' compute and store in a single pass <br>
	 * -) the height of node 'v' in the tree T_1 <br>
	 * -) the cumulative size of all subtrees in 'T_2' rooted at vertex 'v' (along the path from 'v' to 'v0') <br>
	 * 
	 * <br>
	 * It performs the traversal of the contour of the tree, in ccw order.
	 * The traversal starts from the half-edge (v2, v1) and ends at (v0, v1).
	 * (v2, v1) is the half-edge oriented toward v1 corresponding to
	 * the first incoming inner edge incident to v1 (thus of color 1)
	 */
	public void computeHeightT1Fast() {
		/*System.out.print("Fast computation  of the height of nodes in the tree T1...");
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int firstEdge=this.sw.rootEdge.getOpposite().getPrev().getOpposite().getPrev().index; // starting half-edge: inner halfedge of color 1, oriented toward v_1
		int edge01=this.sw.rootEdge.index; // halfedge v01, oriented toward v_1
		int lastEdge=edge01;
		int pEdge=firstEdge; 
		
		//System.out.println("First incoming blue edge (toward at v1): "+firstEdge.getOpposite().getVertex().index+", "+firstEdge.getVertex().index);
		
		int height=0;
		//int count=0;
    	while(pEdge!=lastEdge) { // perform the traversal of the contour of T_1
    		//count++;
    		//System.out.println("pEdge "+count+": "+pEdge.getOpposite().getVertex().index+", "+pEdge.getVertex().index);

    		if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==true) { // an edge of 'T1' visited the first time
    			if(pEdge!=edge01) {
    				height++;
    				
    				int newNode=this.he.getTarget(this.he.getOpposite(pEdge)); // index of the new node encountered during the DFS
    				int ancestor=this.he.getTarget(pEdge); // index of its ancestor node
    				this.setNodeHeight(newNode, 1, height); // store the height of the new (descendant) node in 'T1'
    				
    				//int cumulative=this.getCumulativeSize(ancestor, 0);
    				//this.setCumulativeSize(newNode, 0, cumulative+this.getSubTreeSize(newNode, 0));
    			}
    			//pEdge=pEdge.getPrev();
    			pEdge=this.he.getPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==true) { // a black incoming edge ']'
    			//incomingDegrees=incomingDegrees+']';
    			//pEdge=pEdge.getOpposite().getPrev();
    			// pEdge=this.he.getPrev(this.he.getOpposite(pEdge)); // not efficient
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==false) { // outgoing blue edge
    			//pEdge=pEdge.getOpposite().getPrev();
    			//pEdge=this.he.getPrev(this.he.getOpposite(pEdge));
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==0 && this.sw.isWellOriented[pEdge]==false) { // black outgoing edge
    			//pEdge=pEdge.getOpposite().getPrev();
    			//pEdge=this.he.getPrev(this.he.getOpposite(pEdge));
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==2 && this.sw.isWellOriented[pEdge]==true) { // a blue incoming edge
    			//pEdge=pEdge.getOpposite().getPrev();
    			//pEdge=this.he.getPrev(this.he.getOpposite(pEdge));
    			pEdge=this.he.getOppPrev(pEdge);
    		}
    		else if(this.sw.edgeColor[pEdge]==1 && this.sw.isWellOriented[pEdge]==false) { // ')', an edge of T_0 encountered the second time
    			//treeCode=treeCode+")";
    			height--; 
    			
    			//pEdge=pEdge.getPrev();
    			pEdge=this.he.getPrev(pEdge);
    		}
    	}
    	
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("done");
    	System.out.println(" ("+duration+" seconds)");*/
	}

	/**
	 * Compute and store the 2D Schnyder embedding
	 */
	public void compute2DEmbedding() {
		/*this.coord2D=new Point_2[this.n];
		
		double f=2*this.n-5; // f=2n-5
		
		for(Vertex v: this.sw.polyhedron.vertices) {
			Point_2 p;

			if(v==this.sw.v0)
				p=new Point_2(0.0, 0.);
			else if(v==this.sw.v1)
				p=new Point_2(1., 0.);
			else if(v==this.sw.v2)
				p=new Point_2(0., 1.);
			else {
				double x0, x1, x2;

				// compute the 3 barycentric coordinates
				x0=this.getFaceArea(v.index, 0);
				x1=this.getFaceArea(v.index, 1);
				//x2=this.getFaceArea(v.index, 2);
				
				//System.out.print("v"+v.index+": "+p0+", "+p1+", "+p2+" - ");
				//System.out.print("\t"+v0+", "+v1+", "+v2+"");
				//System.out.println("\t"+x0+", "+x1+", "+x2);
				p=new Point_2(x0/f, x1/f);

			}
			this.coord2D[v.index]=p;
			//System.out.println("v"+v.index+": "+v.getPoint());
		}
		*/
	}

	public String toString() {
		String result="\n";
		
		result=result+"\tP0\tP1\tP2|\tT0\tT1\tT2|\tP0T2\tP1T2\tP1T0\tP2T0\n";
		for(int i=0;i<n;i++) {
			result=result+i+": \t";
			result=result+this.getNodeHeight(i, 0)+"\t"+this.getNodeHeight(i, 1)+"\t"+this.getNodeHeight(i, 2)+" |\t";
			result=result+this.getSubTreeSize(i, 0)+"\t"+this.getSubTreeSize(i, 1)+"\t"+this.getSubTreeSize(i, 2)+" |\t";
			result=result+this.getCumulativeSize(i, 2)+"\t"+this.dataP1T2[i]+"\t"+this.getCumulativeSize(i, 0)+"\t"+this.dataP2T0[i];
			
			result=result+"\n";
		}
		
		result=result+"\n";
		result=result+"\tB0\tB1\tB2|\tR0\tR1\tR2|\tV0\tV1\tV2\n";
		for(int i=0;i<n;i++) {
			result=result+i+": \t";
			result=result+this.getBoundarySize(i, 0)+"\t"+this.getBoundarySize(i, 1)+"\t"+this.getBoundarySize(i, 2)+" |\t";
			result=result+this.getVertexArea(i, 0)+"\t"+this.getVertexArea(i, 1)+"\t"+this.getVertexArea(i, 2)+" |\t";
			
			result=result+"\n";
		}

		return result;
	}

	public void boundaryStat() {
		double separatorSize=Math.sqrt(8.*this.n);
		double size=n/3.;
		int maxSize=0;
		
		int countValid=0;
		for(int i=0;i<n;i++) {
			if(this.getInnerVertexArea(i, 0)>size && this.getInnerVertexArea(i, 0)<2*size && this.getBoundarySize(i, 0)<=separatorSize) {
				countValid++;
				if(this.getVertexArea(i, 0)>maxSize) {
					maxSize=this.getVertexArea(i, 0);
				}
			}
			else if(this.getInnerVertexArea(i, 1)>size && this.getInnerVertexArea(i, 1)<2*size && this.getBoundarySize(i, 1)<=separatorSize) {
				countValid++;
				if(this.getVertexArea(i, 1)>maxSize) {
					maxSize=this.getVertexArea(i, 1);
				}
			}
			else if(this.getInnerVertexArea(i, 2)>size && this.getInnerVertexArea(i, 2)<2*size && this.getBoundarySize(i, 2)<=separatorSize) {
				countValid++;
				if(this.getVertexArea(i, 2)>maxSize) {
					maxSize=this.getVertexArea(i, 2);
				}
			}
		}
		System.out.println("Valid vertices: "+countValid);
		System.out.println("Max size: "+maxSize+", "+(maxSize/(double)n));		
	}
	
	public void minSeparatorStat() {
		double separatorSize=Math.sqrt(8.*this.n);
		double minSize=n/3.; 
		double maxSize=n-minSize;
		
		int countValid=0;
		int R0, R1, R2;
		int P0, P1, P2;
		for(int i=0;i<n;i++) {
			int B0=this.getBoundarySize(i, 0);
			int B1=this.getBoundarySize(i, 1);
			int B2=this.getBoundarySize(i, 2);
			R0=this.getInnerVertexArea(i, 0);
			R1=this.getInnerVertexArea(i, 1);
			R2=this.getInnerVertexArea(i, 2);
			P0=this.getNodeHeight(i, 0);
			P1=this.getNodeHeight(i, 1);
			P2=this.getNodeHeight(i, 2);
			if(B0<=separatorSize || B1<=separatorSize | B2<=separatorSize) {
				double max=Math.max(Math.max(R0, R1), R2);
				if(max<=maxSize) {
					/*System.out.print("v"+i+": "+B0+", "+B1+", "+B2+"\t\t");
					System.out.print(R0+", "+R1+", "+R2+"\t");
					System.out.print(P0+", "+P1+", "+P2);
					System.out.println("\t max="+max/n);
					*/
					countValid++;
				}
			}
		}
		System.out.println("Valid vertices: "+countValid);
		System.out.println("sqrt(8n) "+separatorSize);	
	}

	/**
	 * Computation of the "Shortest balanced separator" (with a prescribed balance ratio)
	 * 
	 * @param balanceRatio  the balance ration allowed for the size of partitions: usually 1/3.
	 * @param size  maximum size allowed for the separator: a constant value, usually between [1., 8.]
	 * @param showResult  choose to print additional information
	 * 
	 * @return
	 */
	public int[] getShortestSeparator(double balanceRatio, double size, boolean showResult) {
		if(this.verbosity>0)
			System.out.print("Computing shortest balanced separator...");
		long startTime=System.nanoTime(), endTime; // for evaluating time performances

		int m=3*this.n-6;
		
		double separatorSize8=Math.sqrt(size*this.n);
		double separatorSize1=Math.sqrt(this.n);
		double separatorSize8m=Math.sqrt(size*m);
		double separatorSize1m=Math.sqrt(m);
		double minSize=n*balanceRatio; 
		double maxSize=n-minSize;
		
		int R0, R1, R2;
		int B0, B1, B2;
		int complement; // size of the partition of the complement of a region: G\R0 (for example)
		
		int bestVertex=-1;
		int partitionSize=0;
		int shortestCycle=Integer.MAX_VALUE; // size of the best cycle separator
		int countValidVertices=0; // LCA not necessary
		boolean isValid=false; // LCA not necessary
		
		for(int i=0;i<n;i++) {
			B0=this.getBoundarySize(i, 0);
			B1=this.getBoundarySize(i, 1);
			B2=this.getBoundarySize(i, 2);

			R0=this.getInnerVertexArea(i, 0);
			R1=this.getInnerVertexArea(i, 1);
			R2=this.getInnerVertexArea(i, 2);
			
			isValid=false;
			if(R0>=minSize && R0<=maxSize) {
				complement=this.n-(R0+B0); // size of G\R0 (including the boundary of R0)

				if(complement>=minSize && complement<=maxSize) {
					if(B0<shortestCycle) {
						shortestCycle=B0;
						bestVertex=i;
						partitionSize=Math.min(R0, complement); // we take the minimum of the two partitions (<= 0.5n)
					}
					if(B0<=separatorSize8) // LCA: not necessary
						isValid=true;
				}
			}
			if(R1>=minSize && R1<=maxSize) {
				complement=this.n-(R1+B1); // size of G\R1 (including the boundary of R1)

				if(complement>=minSize && complement<=maxSize) {
					if(B1<shortestCycle) {
						shortestCycle=B1;
						bestVertex=i;
						partitionSize=Math.min(R1, complement); // we take the minimum of the two partitions (<= 0.5n)
					}
					if(B1<=separatorSize8) // LCA: not necessary
						isValid=true;
				}
			}
			if(R2>=minSize && R2<=maxSize) {
				complement=this.n-(R2+B2); // size of G\R2 (including the boundary of R2)

				if(complement>=minSize && complement<=maxSize) {
					if(B2<shortestCycle) {
						shortestCycle=B2;
						bestVertex=i;
						partitionSize=Math.min(R2, complement); // we take the minimum of the two partitions (<= 0.5n)
					}
					if(B2<=separatorSize8)
						isValid=true;
				}
			}
			
			if(isValid==true) // LCA not necessary
				countValidVertices++;
		}
		
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        if(verbosity>0) {
        	System.out.print("done");
        	System.out.println(" ("+duration+" seconds)");
        }

		if(verbosity>0 && showResult==true) {
			System.out.println("Separator statistics:");
			System.out.println("\tbest vertex: v"+bestVertex);
			System.out.print("\tsqrt("+size+"m)="+((int)separatorSize8m));	
			System.out.print(" sqrt("+size+"n)="+((int)separatorSize8));	
			System.out.print(", sqrt(m)="+((int)(separatorSize1m)));	
			System.out.println(", sqrt(n)="+((int)(separatorSize1)));	
			System.out.println("\tboundary size: "+shortestCycle);	
			System.out.println("\tnormalized boundary size: "+Util.approx((shortestCycle/separatorSize1m),3)+" sqrt(m)");	
			System.out.println("\tnormalized boundary size: "+Util.approx((shortestCycle/separatorSize8),3)+" sqrt(8n)");	
			System.out.println("\tseparator balance: "+partitionSize+" ("+Util.approx((double)partitionSize/n,3)+" n)");
			//System.out.println("Valid vertices: "+countValidVertices+" ("+(double)(countValidVertices)/n+")");
		}
		
		return new int[] {bestVertex, shortestCycle, partitionSize};
	}

	/**
	 * Computation of the "Most balanced short separator" (with a prescribed cycle size)
	 * 
	 * @param size  maximum size allowed for the separator: a constant value, usually between [1., 8.]
	 * @param showResult  choose to print additional information
	 * 
	 * @return
	 */
	public int[] getMostBalancedShortSeparator(double balanceRatio, double size, boolean showResult) {
		double separatorSize=Math.sqrt(size*this.n);
		double separatorSize1=Math.sqrt(this.n);
		double minSize=n*balanceRatio; 
		double maxSize=n-minSize;
		
		int R0, R1, R2;
		int B0, B1, B2;
		int complement; // size of the partition of the complement of a region: G\R0 (for example)
		
		int bestVertex=-1;
		int bestPartition=0;
		int partitionSize=0;
		int cycleSize=Integer.MAX_VALUE; // size of the best cycle separator
		int countValidVertices=0; // LCA not necessary
		boolean isValid=false; // LCA not necessary
		
		for(int i=0;i<n;i++) {
			B0=this.getBoundarySize(i, 0);
			B1=this.getBoundarySize(i, 1);
			B2=this.getBoundarySize(i, 2);

			R0=this.getInnerVertexArea(i, 0);
			R1=this.getInnerVertexArea(i, 1);
			R2=this.getInnerVertexArea(i, 2);
			
			isValid=false;
			if(B0<=separatorSize) { // allow for longer cycle
				complement=this.n-(R0+B0); // size of G\R0 (including the boundary of R0)
				partitionSize=Math.min(R0, complement); // we take the minimum of the two partitions (<= 0.5n)

				if(R0>=minSize && R0<=maxSize && complement>=minSize && complement<=maxSize) {
					if(B0<=separatorSize)
						isValid=true;
					if(partitionSize>bestPartition) {
						cycleSize=B0;
						bestVertex=i;
						bestPartition=partitionSize;
					}
				}
			}
			if(B1<=separatorSize) {
				complement=this.n-(R1+B1); // size of G\R0 (including the boundary of R0)
				partitionSize=Math.min(R1, complement); // we take the minimum of the two partitions (<= 0.5n)

				if(R1>=minSize && R1<=maxSize && complement>=minSize && complement<=maxSize) {
					if(B0<=separatorSize)
						isValid=true;
					if(partitionSize>bestPartition) {
						cycleSize=B1;
						bestVertex=i;
						bestPartition=partitionSize;
					}
				}
			}
			if(B2<=separatorSize) {
				complement=this.n-(R2+B2); // size of G\R0 (including the boundary of R0)
				partitionSize=Math.min(R2, complement); // we take the minimum of the two partitions (<= 0.5n)

				if(R2>=minSize && R2<=maxSize && complement>=minSize && complement<=maxSize) {
					if(B0<=separatorSize)
						isValid=true;
					if(partitionSize>bestPartition) {
						cycleSize=B2;
						bestVertex=i;
						bestPartition=partitionSize;
					}
				}
			}
			
			if(isValid==true) // LCA not necessary
				countValidVertices++;
		}
		
		if(showResult==true) {
			System.out.println("best vertex: v"+bestVertex);
			System.out.print("cycle size: "+cycleSize);	
			System.out.print(" ( sqrt("+size+"n)="+((int)separatorSize));	
			System.out.print(", sqrt("+(size/4.)+"n)="+((int)(separatorSize/(2.)))+")");	
			System.out.println(", sqrt(n)="+((int)(separatorSize1))+")");	
			System.out.println("partition size: "+bestPartition+" ("+(double)bestPartition/n+")");
			System.out.println("Valid vertices: "+countValidVertices+" ("+(double)(countValidVertices)/n+")");
		}
		
		return new int[] {bestVertex, cycleSize, bestPartition};
	}

}
