package Jcg.schnyderwoods;

import java.util.List;

import Jcg.geometry.Point_;
import Jcg.polyhedron.Face;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import Jcg.util.DLinkedList;
import Jcg.util.DListNode;
import arraybased.meshes.MinimalHalfedgeDS;

/**
 * @author Luca Castelli Aleardi (2019, Ecole Polytechnique)
 * 
 * This class allows to compute a Schnyder wood of a planar triangulation.
 * <br>
 * This class makes use of an array-based implementation of the half-edge data structure.
 * As a result, this implementation is faster and uses small memory usage compared to the one
 * based on the standarda half-edge data structure (Polyhedron_3 class).
 */
public class FastSchnyderWood {
	public int verbosity=0;
	
    public boolean[] isWellOriented; // says whether an half-edge is "oriented" toward his target vertex
    public byte[] edgeColor; // stores an integer representing the coloring of an edge

    /** the input mesh */
    public MinimalHalfedgeDS polyhedron;
    
    // definition of the root face and root vertices v0, v1, v2
    /** half-edge (v_0, v_1) oriented toward v_1, assuming the ccw orientation of faces */
    public int rootEdge;
    public int v0, v1, v2; // vertices on the root face (the outer face, in the plane)

    // auxiliary information needed for computing a Schnyder wood
    protected DLinkedList<Integer> outerCycle; // edges defining the outer cycle of the planar map   
    protected boolean[] isChord; // for half-edges, says whether is a chord
    protected boolean[] isOnCutBorder; // say whether a vertex is lying on the cut-border
    
    protected ArrayBasedQueue<DListNode<Integer>> nodes0; // nodes with 0 ingoing red/blue edges 
    protected ArrayBasedQueue<DListNode<Integer>> nodes1; // nodes with 1 ingoing red/blue edges 
    protected ArrayBasedQueue<DListNode<Integer>> nodes2; // nodes with 2 ingoing red/blue edges
    protected ArrayBasedQueue<DListNode<Integer>> nodes3; // nodes with 3 ingoing red/blue edges
    protected ArrayBasedQueue<DListNode<Integer>> nodes4; // nodes with >3 ingoing red/blue edges
    /** store the ingoing degree for each vertex */
    protected int[] ingoing;

    /**
     * Construct the cut-border starting from the root edge (v0, v1)
     * At the beginning the cut-border contains edges (v2, v0) and (v1, v2)
     * Edges are ccw oriented around faces
     */
    public FastSchnyderWood(MinimalHalfedgeDS polyhedron, int rootEdge, int verbosity) {
    	this.verbosity=verbosity;
    	
    	if(verbosity>0)
    		System.out.print("Schnyder wood initialization: ");
    	if(polyhedron==null)
    		throw new Error("error: null polyhedron");
    	if(rootEdge<0 || rootEdge>=polyhedron.sizeOfVertices())
    		throw new Error("error: not valid root edge");
    	
    	this.polyhedron=polyhedron;
    	this.rootEdge=rootEdge;    	
        this.v0=this.polyhedron.getTarget(this.polyhedron.getOpposite(rootEdge));
        this.v1=this.polyhedron.getTarget(rootEdge);
    	
    	this.outerCycle=new DLinkedList<Integer>();

    	this.edgeColor=new byte[this.polyhedron.sizeOfHalfedges()];
    	for(int i=0;i<edgeColor.length;i++)
    		this.edgeColor[i]=-1;
    	
    	this.isChord=new boolean[this.polyhedron.sizeOfHalfedges()];
    	this.isWellOriented=new boolean[this.polyhedron.sizeOfHalfedges()];
    	this.isOnCutBorder=new boolean[this.polyhedron.sizeOfVertices()];
    	
    	int edge10=this.polyhedron.getOpposite(rootEdge);
    	int edge02=this.polyhedron.getNext(edge10);
    	int edge21=this.polyhedron.getNext(edge02);
    	
    	this.v2=this.polyhedron.getTarget(edge02);
    	
    	// setting information concerning the root-edge
    	this.outerCycle.add(this.polyhedron.getOpposite(edge02));	
    	this.outerCycle.add(this.polyhedron.getOpposite(edge21));
    	this.isOnCutBorder[v1]=true;
    	this.isOnCutBorder[v0]=true; // LCA newly added
    	this.isOnCutBorder[v2]=true; // LCA newly added
    	
    	// set the orientation of the root edge
    	this.isWellOriented[rootEdge]=false;
    	this.isWellOriented[this.polyhedron.getOpposite(rootEdge)]=true;
    	// set the color of the root edge
    	this.edgeColor[rootEdge]=0;
    	this.edgeColor[this.polyhedron.getOpposite(rootEdge)]=0;
    	
    	if(verbosity>0) {
    		    	System.out.print("\t root face (v"+v0+", v"+v1+", v"+v2+")");
    		    	System.out.println("\t root edge e"+rootEdge+" (v"+this.polyhedron.getTarget(this.polyhedron.getOpposite(rootEdge))+", v"+this.polyhedron.getTarget(rootEdge)+")");
    	}
    	
    	this.nodes0=new ArrayBasedQueue<DListNode<Integer>>(this.polyhedron.sizeOfVertices());
    	this.nodes1=new ArrayBasedQueue<DListNode<Integer>>(this.polyhedron.sizeOfVertices());
    	this.nodes2=new ArrayBasedQueue<DListNode<Integer>>(this.polyhedron.sizeOfVertices());
    	this.nodes3=new ArrayBasedQueue<DListNode<Integer>>(this.polyhedron.sizeOfVertices());
    	this.nodes4=new ArrayBasedQueue<DListNode<Integer>>(this.polyhedron.sizeOfVertices());
    	this.ingoing=new int[this.polyhedron.sizeOfVertices()];
    	for(int i=0;i<this.ingoing.length;i++) 
    		this.ingoing[i]=-1;
   }
    
    /**
     * It removes a vertex from the cut-border (vertex conquest)
     * It updates the cut-border, assigning color and orientation to edges
     * 
     * @param node The node in the list storing the given half-edge (whose vertex will be removed)
     * @return result return the list node containing the preceding halfedge (on the boundary cycle)
     */
    public DListNode<Integer> vertexRemoval(DListNode<Integer> node) {
    	if(node==null || this.outerCycle.isEmpty()) {
    		System.out.println("no more vertex to remove ");
    		return null;
    	}
    	
    	Integer rightEdge=node.getElement(); // the vertex v of rightEdge will be removed
    	if(rightEdge==null) {
    		throw new Error("null reference: rightEdge");
    	}
    	if(this.polyhedron.getTarget(rightEdge)==this.v0) // vertex v0 cannot be removed
    		return node.getNext();
    	
    	Integer leftEdge=node.getPrev().getElement();
    	if(leftEdge==null) {
    		throw new Error("null reference: leftEdge");
    	}
    	
    	if(this.hasIncidentChords(node)==true)
    		return node.getNext();
    	
    	// process left and right edges incident to v on the cut-border
    	this.setOutgoingEdge1(rightEdge);
    	this.setOutgoingEdge0(leftEdge);
    	
    	// for computing balanced Schnyder woods
    	int leftNode=this.polyhedron.getTarget(leftEdge);
    	int rightNode=this.polyhedron.getTarget(this.polyhedron.getOpposite(rightEdge)); // LCA getSource
    	
    	// v is not anymore on the cut-border
    	this.isOnCutBorder[this.polyhedron.getTarget(rightEdge)]=false;
    	
    	// add new edges to the cut-border
    	//this.addToCutBorder(rightEdge.getPrev().getOpposite(), i-1);
    	int b=this.polyhedron.getOpposite(this.polyhedron.getPrev(rightEdge));
    	this.setToCutBorder(b);
    	node.setElement(b);
    	this.addToQueue(node); // add to the queue the (new) rightmost vertex on the path, from v_l to v_r

    	DListNode<Integer> previousNode=node.getPrev(); // the boundary edge preceding
    	DListNode<Integer> nextNode=node.getNext(); // the boundary edge following
    	
    	this.addToQueue(nextNode); // for balanced Schnyder woods: incoming blue edge (right)
   	
    	int pEdge=this.polyhedron.getOpposite(this.polyhedron.getNext(rightEdge));
    	while(pEdge!=this.polyhedron.getOpposite(leftEdge)) {
    		this.setToCutBorder(this.polyhedron.getOpposite(this.polyhedron.getPrev(pEdge)));
    		this.outerCycle.insertAfter(previousNode, this.polyhedron.getOpposite(this.polyhedron.getPrev(pEdge)));
    		this.setIngoingEdge2(pEdge);
    		
        	this.addToQueue(previousNode.getNext()); // for balanced Schnyder woods

    		pEdge=this.polyhedron.getOpposite(this.polyhedron.getNext(pEdge));
    	}
    	DListNode<Integer> result=previousNode.getNext();
    	this.outerCycle.delete(previousNode);
    	
    	return result;
    }

    /**
     * Add each vertex to a queue, depending on its ingoing degree
     */
    protected void addToQueue(DListNode<Integer> node) {
    	if(node==null || node.getElement()==null)
    		return;
    	
    	int e=node.getElement();
    	if(this.polyhedron.getTarget(e)==this.v0 || this.polyhedron.getTarget(e)==this.v1) // LCA: this is instruction is redundant (just for debugging)
    		return;
    	
    	int vertex=this.polyhedron.getTarget(e);
    	this.ingoing[vertex]++;
    	
    	//System.out.println("\t\t\tadding v"+vertex+", ingoing degree="+this.ingoing[vertex]);
    	
    	if(this.ingoing[vertex]==0) {
    		this.nodes0.add(node);
    	}
    	else if(this.ingoing[vertex]==1) {
    		this.nodes1.add(node);
    	}
    	else if(this.ingoing[vertex]==2) {
    		this.nodes2.add(node);
    	}
    	else if(this.ingoing[vertex]==3) {
    		this.nodes3.add(node);
    	}
    	else
    		this.nodes4.add(node);
    }
    
    /**
     * Heuristic computation of a balanced Schnyder wood
     */
    public void performBalancedTraversal() {
   		//System.out.print("Fast computation of a balanced Schnyder wood (for planar triangulations)...");
    	//System.out.println("Cut border: \n"+this.toString());
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	//if(this.verbosity>0)
    	//	System.out.println("First phase");
    	
    	DListNode<Integer> node=this.outerCycle.getFirst().getNext();
    	if(node==null) {
    		throw new Error("error null reference: first boundary node not defined");
    	}
    	int count=0;
    	while(this.outerCycle.size()>1 && count<1) {
    		//System.out.println("\nCut border: \n"+this.toString());
    		//System.out.println("Processing vertex "+node.getElement().getVertex().index);
    		node=this.vertexRemoval(node);
    		count++;
    	}

    	//if(this.verbosity>0)
    	//	System.out.println("Intermediate phase");
    	
    	node=this.outerCycle.getFirst().getNext();
    	/*if(node==null) {
    		throw new Error("error null reference: first boundary node not defined");
    	}*/
    	//count=0;
    	//int countRemoved=0;
    	while((this.nodes4.isEmpty()==false || this.nodes3.isEmpty()==false || this.nodes2.isEmpty()==false || this.nodes1.isEmpty()==false || this.nodes0.isEmpty()==false)) {
    		//this.printCutBorderInfo();
    		
    		if(this.nodes4.isEmpty()==false)
    			node=this.nodes4.poll();
    		else if(this.nodes3.isEmpty()==false)
    			node=this.nodes3.poll();
    		else if(this.nodes2.isEmpty()==false)
    			node=this.nodes2.poll();
    		else if(this.nodes1.isEmpty()==false)
    			node=this.nodes1.poll();
    		else
    			node=this.nodes0.poll();
    		
    		if(node!=null && node.getElement()!=null) {
    			int e=node.getElement();
    			
    			if(this.isOnCutBorder[this.polyhedron.getTarget(e)]==true) {
    				this.vertexRemoval(node);
    				//countRemoved++;
    			}
    		}
    	}
    	/*if(this.verbosity>0) {
    		System.out.println("Removed: "+countRemoved);
    		//this.printCutBorderInfo();
    	}

    	if(this.verbosity>0)
    		System.out.println("--- Final phase ---");
    	*/
    		
    	if(this.outerCycle.size()>1)
    		throw new Error("Error: empty queues");
    	
    	node=this.outerCycle.getFirst().getNext();
    	/*if(node==null) {
    		throw new Error("error null reference: first boundary node not defined");
    	}*/
    	//countRemoved=0;
    	while(this.outerCycle.size()>1) {
    		//System.out.println("\nCut border: \n"+this.toString());
    		//System.out.println("Processing vertex "+node.getElement().getVertex().index);
    		//node=this.vertexRemoval(node);
    		//countRemoved++;
    	}
    	//if(this.verbosity>0)
    	//	System.out.println("Removed in final phase: "+countRemoved);
    	
    	// set the color of the root edge, which is of color 0 (oriented toward v0)
    	this.edgeColor[rootEdge]=0;
    	this.edgeColor[this.polyhedron.getOpposite(rootEdge)]=0;
        
    	/*endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");*/
    	//System.out.println("Triangle removals: "+count);
    }

    /**
     * Check whether a vertex has incident chords
     * In that case it cannot be removed from the cut-border
     */
    protected boolean hasIncidentChords(DListNode<Integer> node) {
    	int rightEdge=node.getElement();
    	int leftEdge=node.getPrev().getElement();

    	int pEdge=this.polyhedron.getOpposite(this.polyhedron.getNext(rightEdge));
    	int v;
    	while(pEdge!=this.polyhedron.getOpposite(leftEdge)) {
    		v=this.polyhedron.getTarget(this.polyhedron.getOpposite(pEdge));
    		if(this.isOnCutBorder[v]==true)
    			return true;
    		pEdge=this.polyhedron.getOpposite(this.polyhedron.getNext(pEdge));
    	}
    	return false;
    }

    protected void setIngoingEdge2(int e) {
    	this.edgeColor[e]=2;
    	this.edgeColor[this.polyhedron.getOpposite(e)]=2;
    	
    	this.isWellOriented[e]=true;
    	this.isWellOriented[this.polyhedron.getOpposite(e)]=false;
    }

    protected void setOutgoingEdge1(int e) {
    	this.edgeColor[e]=1;
    	this.edgeColor[this.polyhedron.getOpposite(e)]=1;
    	
    	this.isWellOriented[e]=false;
    	this.isWellOriented[this.polyhedron.getOpposite(e)]=true;
    }

    protected void setOutgoingEdge0(int e) {
    	this.edgeColor[e]=0;
    	this.edgeColor[this.polyhedron.getOpposite(e)]=0;
    	
    	this.isWellOriented[e]=true;
    	this.isWellOriented[this.polyhedron.getOpposite(e)]=false;
    }

    /**
     * Add a half-edge to the cut-border, at a given position (in the cut-border)
     * Update all concerned information (colors, existent chords, boundary vertices, ...)
     */
    public void addToCutBorder(Integer e, DListNode<Integer> node) {
    	//System.out.println("edge added to cut-border");
    	if(e==null)
    		throw new Error("halfedge not defined");
    	if(node==null)
    		throw new Error("error null reference: list node");
    	
    	this.outerCycle.insertBefore(node, e);
    	this.setToCutBorder(e);
    }

    /**
     * Add a half-edge to the cut-border, at a given position (in the cut-border)
     * Update all concerned information (colors, existent chords, boundary vertices, ...)
     */
    public void setToCutBorder(int e) {
    	this.edgeColor[e]=3;
    	this.edgeColor[this.polyhedron.getOpposite(e)]=3;
    	
    	this.isChord[e]=false;
    	this.isChord[this.polyhedron.getOpposite(e)]=false;
    	
    	this.isOnCutBorder[this.polyhedron.getTarget(e)]=true;
    	this.isOnCutBorder[this.polyhedron.getTarget(this.polyhedron.getOpposite(e))]=true;
    }

    /**
     * Perform all steps of the vertex conquest, computing the minimal Schnyder wood
     */
    public void performTraversal() {
    	System.out.print("Fast computation of a Schnyder wood (for planar triangulations)...");
    	//System.out.println("Cut border: \n"+this.toString());
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	DListNode<Integer> node=this.outerCycle.getFirst().getNext();
    	if(node==null) {
    		throw new Error("error null reference: first boundary node not defined");
    	}
    	while(this.outerCycle.size()>1) {
    		//System.out.println("\nCut border: \n"+this.toString());
    		//System.out.println("Processing vertex "+node.getElement().getVertex().index);
    		node=this.vertexRemoval(node);
    	}
    	// set the color of the root edge, which is of color 0 (oriented toward v0)
    	this.edgeColor[rootEdge]=0;
    	this.edgeColor[this.polyhedron.getOpposite(rootEdge)]=0;
        
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("done");
    	System.out.println(" ("+duration+" seconds)");
    	//System.out.println("Triangle removals: "+count);
    }
        
    /**
     * Return an array of indices representing the original vertex ordering
     */
    /*public int[] getOriginalVertexOrdering() {
    	int[] result=new int[this.polyhedron.sizeOfVertices()];
    	
    	int i=0;
    	for(Object o: this.polyhedron.vertices) {
    		Vertex<Point_> v=(Vertex<Point_>)o;
    		result[i]=v.index;
    		i++;
    	}
    	return result;    	
    }*/

	/**
	 * Return the first (half)edge oriented toward vertex v and having color 0 (red)
	 * Edges (v0, v1) and (v2, v0) are assumed to be red colored
	 * Return null, if such an edge does not exist
	 */
	/*public int getFirstIncomingRedEdge(int v) {
		if(v==this.v0) {
			return this.polyhedron.getOpposite(rootEdge); // half-edge (v1, v0), oriented toward v0
		}
		if(v==this.v1) // no incoming red edges
			return -1;
		if(v==this.v2) // no incoming red edges
			return -1;
		
    	// visit all (half)-edges incident to v, and the return the first incoming edge of color 0
		Halfedge<Point_> e=v.getHalfedge();
		if(this.edgeColor[e.index]==0 && this.isWellOriented[e.index]==true && this.edgeColor[e.getNext().index]==1)
			return e;
    	Halfedge<Point_> pEdge=e.getOpposite().getPrev();
    	while(pEdge!=e) { // turn around vertex v
    		if(this.edgeColor[pEdge.index]==0 && this.isWellOriented[pEdge.index]==true && this.edgeColor[pEdge.getNext().index]==1)
    			return pEdge;
    		pEdge=pEdge.getOpposite().getPrev();
    	}
    	return null;
	}*/

	/**
	 * Return the next edge of color 0, after edge e (turning in ccw direction), oriented toward its target vertex (denoted by v).
	 * Half-edge e is assumed to have v as target vertex
	 * Edges (v0, v1) and (v2, v0) are assumed to be red colored (and oriented toward v0)
	 * Return null, if such an edge does not exist
	 */
	/*public Halfedge<Point_> getNextIncomingRedEdge(Halfedge<Point_> e) {
		if(e==null) return null; 
		Halfedge<Point_> e20=this.rootEdge.getOpposite().getPrev(); // halfedge e=(v2, v0), having v0 as target
		if(e==e20) {
			return null; // e20 is the last ingoing edge (of color 0) incident to v0
		}
		
    	Halfedge<Point_> pEdge=e.getOpposite().getPrev();
    	if(this.edgeColor[pEdge.index]==0 && this.isWellOriented[pEdge.index]==true)
    			return pEdge;
    	return null;
	}*/
	
	/**
	 * Check whether the face incident to the edge 'e' is CCW oriented <br>
	 * 
	 * Warning: it works only for triangular faces <br>
	 * 
	 * Remark: the faces are assumed to be ccw oriented in the embedding
	 */
	/*public boolean isCCWOriented(int e) {
		Face root=this.rootEdge.getOpposite().getFace(); // exterior face
		if(e.getFace()==root)
			return false;
		
		if(this.isWellOriented[e]==false)
			return false;
		if(this.isWellOriented[this.polyhedron.getNext(e)]==false)
			return false;
		if(this.isWellOriented[this.polyhedron.getPrev(e)]==false)
			return false;
		
		return true;
	}*/
	
	/**
	 * Check whether the triangle face incident to the half-edge 'e' is CW oriented <br>
	 * 
	 * Warning: it works only for triangular faces <br>
	 * 
	 * Remark: the faces are assumed to be ccw oriented in the embedding
	 */
	/*public boolean isCWOriented(int e) {
		Face root=this.rootEdge.getOpposite().getFace(); // exterior face
		if(e.getFace()==root)
			return false;
		
		if(this.isWellOriented[e]==true)
			return false;
		if(this.isWellOriented[this.polyhedron.getNext(e)]==true)
			return false;
		if(this.isWellOriented[[this.polyhedron.getPrev(e)]==true)
			return false;
		
		return true;
	}*/
		
    //----------------------------------------------
    //--- Methods for visualizing the cut-border ---
    //----------------------------------------------

}
