package Jcg.simplification;

import java.util.ArrayList;
import java.util.PriorityQueue;

import Jcg.geometry.Point_3;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;

public class EdgeContraction extends MeshSimplification {

	public EdgeContraction(Polyhedron_3<Point_3> polyhedron3d) {
		super(polyhedron3d);
	}

	@Override
	public void simplify(double ratio) {
		this.polyhedron3D.resetMeshIndices();
		
		Halfedge<Point_3> e;
		Vertex<Point_3> u;
		Point_3 p = new Point_3();

		int nOfEdges = polyhedron3D.halfedges.size()/2;
		int nOfSimplifications = (int)(nOfEdges*ratio);

		//create priority queue storing half-edges
		PriorityQueue<Halfedge<Point_3>> q = new PriorityQueue<Halfedge<Point_3>>(nOfEdges, new HalfEdgeComparator());
		for(Halfedge<Point_3> edge: this.polyhedron3D.halfedges) {
			if(edge.getVertex().index>edge.getOpposite().getVertex().index)
				q.offer(edge); // insert only one of the half-edges
		}

		ArrayList<Halfedge<Point_3>> list;

		int i=0;
		while(i<nOfSimplifications && q.isEmpty()==false){

			//Take halfedge to be analyzed (and remove it's opposite halfedge, since it corresponds to the same pair of vertex)
			e = q.poll();
			q.remove(e.opposite);
			//Makes list of internal halfedges that should be removed from queue in case contraction takes place
			ArrayList<Halfedge<Point_3>> listToRemove = new ArrayList<Halfedge<Point_3>>();
			listToRemove.add(e.next);
			listToRemove.add(e.next.next);
			listToRemove.add(e.opposite.next);
			listToRemove.add(e.opposite.next.next);

			u = e.opposite.vertex;
			
			if(isCollapsible(e)) { //Check whether 'e' can be collapsed

				// Barycenter considering u, v and all neighbors
				ArrayList<Vertex<Point_3>> neighborsV = getNeighbors(e);
				ArrayList<Vertex<Point_3>> neighborsU = getNeighbors(e.opposite);
				int nV = neighborsV.size();
				int nU = neighborsU.size();
				//get points u,v and neighbors
				Point_3[] points = new Point_3[nV + nU + 2];
				for(int k=0; k<nV; k++)
					points[k] = neighborsV.get(k).getPoint();
				for(int k=0; k<nU; k++)
					points[nV+k] = neighborsU.get(k).getPoint();
				points[nV+nU] = e.getVertex().getPoint();
				points[nV+nU+1] = u.getPoint();
				//set coefficients for linear combination (mean)
				Number[] coefficients = new Number[nV + nU + 2];
				for(int k=0; k<(nU+nV+2); k++)
					coefficients[k] = 1.f/(nU+nV+2.f);
				p = Point_3.linearCombination(points, coefficients);

				//Collapse edge and adjust point
				this.polyhedron3D.edgeCollapse(e);
				u.setPoint(p);

				//remove and re-insert neighbors into queue
				list = getIncidentHalfedges(u);
				for(Halfedge<Point_3> h : list){
					q.remove(h);
					q.offer(h);
				}
				//removes from queue halfedges that no longer exist
				for(Halfedge<Point_3> h : listToRemove)
					q.remove(h);
				
				i++;
			}
		}

	}

	boolean isCollapsible(Halfedge<Point_3> h){
		ArrayList<Vertex<Point_3>> neighborsV = getNeighbors(h);
		ArrayList<Vertex<Point_3>> neighborsU = getNeighbors(h.opposite);

		return !hasIllegalSharedVertexNeighborNumber(neighborsV, neighborsU);
	}

	private ArrayList<Vertex<Point_3>> getNeighbors(Halfedge<Point_3> h){
		ArrayList<Vertex<Point_3>> l = new ArrayList<Vertex<Point_3>>();
		Halfedge<Point_3> initialH = h.next;
		Halfedge<Point_3> currentH = initialH;

		do{
			l.add(currentH.vertex);
			currentH = currentH.opposite.next;
		}while(!l.contains(currentH.vertex));

		return l;
	} 

	private boolean hasIllegalSharedVertexNeighborNumber(ArrayList<Vertex<Point_3>> a, ArrayList<Vertex<Point_3>> b){
		int n = 0;
		for(Vertex<Point_3> v : a){
			if(b.contains(v))
				n++;
			if(n>2) return true;
		}
		return false;
	}

	private ArrayList<Halfedge<Point_3>> getIncidentHalfedges(Vertex<Point_3> v){
		ArrayList<Halfedge<Point_3>> l = new ArrayList<Halfedge<Point_3>>();
		Halfedge<Point_3> initialH = v.getHalfedge();
		Halfedge<Point_3> currentH = initialH;
		do{
			l.add(currentH);
			currentH = currentH.next.opposite;
		}while(!l.contains(currentH));

		initialH = v.getHalfedge().opposite;
		currentH = initialH;
		do{
			l.add(currentH);
			currentH = currentH.opposite.next;
		}while(!l.contains(currentH));

		return l;
	} 

}



