package Jcg.subdivision;

import Jcg.geometry.*;
import Jcg.polyhedron.*;

import java.util.*;

/**
 * Perform the Loop subdivision scheme
 * Only for triangular meshes (without boundary)
 * 
 * @author Luca Castelli Aleardi (INF555, 2012)
 *
 */
public class CatmullClarkSubdivision extends MeshSubdivision {
  
  HashSet<Vertex<Point_3>> originalVertices; // store the references of the original vertices
  HashMap<Face, Point_3> faceCentroids;
  HashMap<Vertex<Point_3>, Point_3> midPoints;
  
  public CatmullClarkSubdivision(Polyhedron_3<Point_3> polyhedron) {
    super(polyhedron);
  }
  
  /**
   * The main method performing the subdivision process
   * To be implemented
   */
  public void subdivide() {
    originalVertices = new HashSet<Vertex<Point_3>>();
    for (Vertex v: this.polyhedron3D.vertices) {
      originalVertices.add(v);
    }
    faceCentroids = computeAllCentroids();
    midPoints = computeAllmidPoints();
    splitEdges(computeEdgePoints());
 
    Point_3[] newPositions = this.computeNewVertexLocations();
    for (int i = 0; i < this.polyhedron3D.vertices.size(); i++) {
      this.polyhedron3D.vertices.get(i).setPoint(newPositions[i]);
    }
    
    Stack<Face<Point_3>> facets = new Stack<Face<Point_3>>();
    for (Face<Point_3> f : faceCentroids.keySet()) {
      facets.add(f);
    }
    while (!facets.isEmpty()) {
      Face<Point_3> facet = facets.pop();
      subdivideFace(facet);
    }
  }
  
  private HashMap<Vertex<Point_3>, Point_3> computeAllmidPoints() {
    HashMap<Vertex<Point_3>, Point_3> result = new HashMap<Vertex<Point_3>, Point_3>();
    for (Vertex<Point_3> v: this.polyhedron3D.vertices) {
      ArrayList<Vertex<Point_3>> neighbours = this.computeVertexNeighbours(v);
      int D = neighbours.size();
      Number[] coeffs = new Number[D];
      Point_3[] points = new Point_3[D];
      for (int i = 0; i < D; i++) {
        coeffs[i] = 1.0/D;
        Point_3[] interPoints = {v.getPoint(), neighbours.get(i).getPoint()};
        Number[] intercoeff = {0.5, 0.5};
        points[i] = Point_3.linearCombination(interPoints, intercoeff); 
      }
      result.put(v, Point_3.linearCombination(points, coeffs));
    }
    return result;
  }

  public HashMap<Face, Point_3> computeAllCentroids() {
    HashMap<Face, Point_3> centroids = new HashMap<Face, Point_3>();
    for (Face f : this.polyhedron3D.facets) {
        centroids.put(f, computeFaceCentroid(f));
    }
    return centroids;
  }
  
  public Point_3 computeFaceCentroid(Face f) {
    int d = f.degree();
    int[] vertices = f.getVertexIndices(this.polyhedron3D);
    Point_3[] points = new Point_3[d];
    Number[] coeff = new Number[d];
    for (int i = 0; i < vertices.length; i++) {
      points[i] = this.polyhedron3D.vertices.get(vertices[i]).getPoint();
      coeff[i] = (1.0/d);
    }
    return Point_3.linearCombination(points, coeff);
  }

  /**
   * Splits all edges by inserting a new vertex
   */
  public void splitEdges(HashMap<Halfedge<Point_3>, Point_3> edgePoints) {
    System.out.print("Splitting edges...");
    for (Halfedge<Point_3> h : edgePoints.keySet()) {
      this.polyhedron3D.splitEdge(h, edgePoints.get(h));
    }
    System.out.println("done");
  }
  
  /**
   * Perform the subdivision of a face into 4 triangular sub-faces
   * Edges must already be split: the face has degree 3+3
   */
  public void subdivideFace(Face<Point_3> f) {
    Halfedge h1 = f.getEdge();
    if (originalVertices.contains(h1.vertex)) {
      h1 = h1.next;
    }
    Halfedge hLast = h1.prev.prev;
    f.setEdge(hLast);
    Halfedge h = h1.next.next;
    //Adding a new vertex
    Vertex newVertex = new Vertex();
    newVertex.setPoint(this.faceCentroids.get(f));
    this.polyhedron3D.vertices.add(newVertex);
    
    //New Face 
    Face newFace = new Face();
    newFace.setEdge(h1);
    this.polyhedron3D.facets.add(newFace);
    
    //New halfedges
    Halfedge newH1 = new Halfedge();
    Halfedge newH2 = new Halfedge();
    newH1.setVertex(newVertex);
    newH2.setVertex(hLast.getVertex());
    newH1.setOpposite(newH2);
    newH2.setOpposite(newH1);
    newH2.setNext(hLast.next);
    hLast.next.setPrev(newH2);
    newH1.setPrev(hLast);
    hLast.setNext(newH1);
    newH1.setFace(f);
    newH2.setFace(newFace);
    this.polyhedron3D.halfedges.add(newH1);
    this.polyhedron3D.halfedges.add(newH2);
    
    Halfedge newH3 = new Halfedge();
    Halfedge newH4 = new Halfedge();
    newH3.setVertex(newVertex);
    newH4.setVertex(h1.getVertex());
    newH3.setOpposite(newH4);
    newH4.setOpposite(newH3);
    newH4.setNext(h1.next);
    h1.next.setPrev(newH4);
    newH3.setPrev(h1);
    h1.setNext(newH3);
    newH3.setFace(newFace);
    newH4.setFace(f);
    newH1.setNext(newH4);
    newH2.setPrev(newH3);
    newH3.setNext(newH2);
    newH4.setPrev(newH1);
    this.polyhedron3D.halfedges.add(newH3);
    this.polyhedron3D.halfedges.add(newH4);
   
    h1.setFace(newFace);
    h1.prev.setFace(newFace);
    
    newVertex.setEdge(newH1);
    
    while (!h.equals(hLast)) {
      Halfedge temp = h.next.next;
      this.polyhedron3D.splitFacet(h, hLast.next);
      h = temp;
    }
    // this.polyhedron3D.splitFacet(h.next.next.next.next, h.next.next);
   /* h1 = this.polyhedron3D.splitFacet(h1, h1.next.next);
    h1 = this.polyhedron3D.splitFacet(h1, h1.next.next);
    h1 = this.polyhedron3D.splitFacet(h1, h1.next.next);
    */
  }

  /**
   * Compute a new edge point (given the half-edge h)
   */
  public Point_3 computeEdgePoint(Halfedge<Point_3> h) {
    Point_3 v0 = h.getVertex().getPoint();
    Point_3 v1 = h.getOpposite().getVertex().getPoint();
    Point_3 f1 = this.faceCentroids.get(h.face);
    Point_3 f2 = this.faceCentroids.get(h.opposite.face);
    Number[] coeff = {0.25, 0.25, 0.25, 0.25};
    Point_3[] points = {v0, v1 , f1, f2};
    return Point_3.linearCombination(points, coeff);
  }

  /**
   * Compute all new edge points and store the result in an HashMap
   */
  public HashMap<Halfedge<Point_3>, Point_3> computeEdgePoints() {
    HashMap<Halfedge<Point_3>, Point_3> result = 
        new HashMap<Halfedge<Point_3>, Point_3>();
        for (Halfedge<Point_3> h : this.polyhedron3D.halfedges) {
          if (h.opposite.tag == 0) {
            result.put(h, computeEdgePoint(h));
            h.tag = 1;
          }
        }
        return result;
  }
  
  /**
   * Compute the new coordinates for a vertex (already existing in the initial mesh)
   */
  public Point_3 computeNewVertexLocation(Vertex<Point_3> v) {
   Set<Face> incidentFaces = computeIncidentFaces(v);
   // Calcul de F:
   Number[] coeffs = new Number[incidentFaces.size()];
   Point_3[] points = new Point_3[incidentFaces.size()];
   int i = 0; 
   for (Face f: incidentFaces) {
     points[i] = this.faceCentroids.get(f);
     coeffs[i] = 1.0 / incidentFaces.size();
     i++;
   }
   Point_3 F = Point_3.linearCombination(points, coeffs);
   Point_3 E = this.midPoints.get(v);
   int d = this.computeVertexNeighbours(v).size();
   Number[] realcoeffs = {(d-3.0)/d, 2.0 /d, 1.0 / d};
   Point_3[] realPoints = {v.getPoint(), E, F};
   return Point_3.linearCombination(realPoints, realcoeffs);
  }

  private Set<Face> computeIncidentFaces(Vertex<Point_3> v) {
    Set<Face> faces = new HashSet<Face>();
    Halfedge h = v.getHalfedge();
    faces.add(h.getFace());
    Halfedge other = h.next.opposite;
    while (!other.equals(h)) {
      faces.add(other.getFace());
      other = other.next.opposite;
    }
    return faces;
  }

  /**
   * Compute the new coordinates for all vertices of the initial mesh
   */
  public Point_3[] computeNewVertexLocations() {
    int N = this.polyhedron3D.vertices.size();
    Point_3[] newPoints = new Point_3[N];
    for (int i = 0; i < N; i++) {
      Vertex v = this.polyhedron3D.vertices.get(i);
        if (this.originalVertices.contains(v)) {
          newPoints[i] = this.computeNewVertexLocation(v);
        } else {
          newPoints[i] = (Point_3) v.getPoint();
        }
    }
    return newPoints;
  }
  
  public ArrayList<Vertex<Point_3>> computeVertexNeighbours(Vertex<Point_3> v) {
    ArrayList<Vertex<Point_3>> neighbours = new ArrayList<Vertex<Point_3>>();
    Halfedge h = v.getHalfedge();
    h = h.opposite;
    neighbours.add(h.vertex);
    Halfedge g = v.getHalfedge().next;
    while (!g.equals(h)) {
      neighbours.add(g.vertex);
      g = g.opposite.next;
    }
    return neighbours;
  }
  
  
}
