package Jcg.subdivision;
import Jcg.geometry.*;
import Jcg.polyhedron.*;

import java.util.*;

/**
 * Perform the Loop subdivision scheme
 * Only for triangular meshes (without boundary)
 * 
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2012)
 *
 */
public class LoopSubdivision extends MeshSubdivision {
	
	HashSet<Vertex<Point_3>> originalVertices; // store the references of the original vertices
	
	public LoopSubdivision(Polyhedron_3<Point_3> polyhedron) {
		super(polyhedron);
	}
	
	/**
	 * The main method performing the subdivision process
	 * To be implemented
	 */
	public void subdivide() {		
		// useful for distinguishing between original vertices and the new ones (added by the subdivision scheme)
		originalVertices=new HashSet<Vertex<Point_3>>();
		for(Vertex<Point_3> v: this.polyhedron3D.vertices) {
			this.originalVertices.add(v);
		}
		
		// store new vertex locations for the vertices of the initial mesh
		Point_3[] newVertexLocations=this.computeNewVertexLocations();
		
		HashMap<Halfedge<Point_3>, Point_3> edgePoints=this.computeEdgePoints();
		splitEdges(edgePoints); // perform the split of all (original) edges
		
		// remark: cannot access the Arraylist polyhedron.facets, because of concurrency
		LinkedList<Face<Point_3>> facesToSubdivide=new LinkedList<Face<Point_3>>();
		for(Face<Point_3> f: this.polyhedron3D.facets) {
			facesToSubdivide.add(f);
		}
		for(Face<Point_3> f: facesToSubdivide)
			subdivideFace(f);
		
		int i=0;
		for(Vertex<Point_3> v: this.polyhedron3D.vertices) {
			v.setPoint(newVertexLocations[i]);
			i++;
			if(i>=newVertexLocations.length)
				return;
		}
	}

	/**
	 * Splits all edges by inserting a new vertex
	 */
	public void splitEdges(HashMap<Halfedge<Point_3>, Point_3> edgePoints) {
		//System.out.print("Splitting edges...");
		// set of half-edges which have been already split
		LinkedList<Halfedge<Point_3>> edgesToSplit=new LinkedList<Halfedge<Point_3>>();
		for(Halfedge<Point_3> h: this.polyhedron3D.halfedges) {
			h.tag=1; // set the edge to be split;
			edgesToSplit.add(h);
		}
		
		while(edgesToSplit.size()>0) {
			Halfedge<Point_3> h=edgesToSplit.pop();
			if(h.tag==1) { // the halfedge, and its opposite, must be split
				Point_3 midPoint=edgePoints.get(h);
				this.polyhedron3D.splitEdge(h, midPoint);
				
				h.tag=0;
				h.opposite.tag=0;
			}
		}
		System.out.println("done");
	}
	
	/**
	 * Perform the subdivision of a face into 4 triangular sub-faces
	 * Edges must already be split: the face has degree 3+3
	 */
	public void subdivideFace(Face<Point_3> f) {
		//throw new Error("To be completed");
		if(f==null || f.degree()<4)
			return;
		Halfedge<Point_3> h1=f.getEdge();
		if(originalVertices.contains(h1.getVertex())==true) // the half-edge must point to a "new midpoint" vertex
			h1=h1.getNext();
		
		Halfedge<Point_3> g1=h1.getNext().getNext();
		Halfedge<Point_3> h2=this.polyhedron3D.splitFacet(h1, g1);
		Halfedge<Point_3> g2=h2.getNext().getNext();
		Halfedge<Point_3> g3=this.polyhedron3D.splitFacet(h2, g2);
		this.polyhedron3D.splitFacet(h1, g3);
	}

	/**
	 * Compute a new edge point (given the half-edge h)
	 */
	public Point_3 computeEdgePoint(Halfedge<Point_3> h) {
		//throw new Error("To be completed");
		
		// retrieve the coordinates of the four points incident to the two faces sharing h
		Point_3 v=h.getVertex().getPoint();
		Point_3 u=h.getOpposite().getVertex().getPoint();
		Point_3 left=h.getNext().getVertex().getPoint();
		Point_3 right=h.getOpposite().getNext().getVertex().getPoint();
		
		Point_3[] points={u, v, left, right};
		Double[] coefficients={3./8, 3./8, 1./8, 1./8};
		Point_3 result=Point_3.linearCombination(points, coefficients);
		
		return result;
	}

	/**
	 * Compute all new edge points and store the result in an HashMap
	 */
	public HashMap<Halfedge<Point_3>, Point_3> computeEdgePoints() {
		//throw new Error("To be completed");
		
		// store the new edge points in an hash table
		HashMap<Halfedge<Point_3>, Point_3> result=new HashMap<Halfedge<Point_3>, Point_3>();
		for(Halfedge<Point_3> h: this.polyhedron3D.halfedges) {
			//if(result.containsKey(h.getOpposite())==false) // only for reducing the number of entries in the table
				result.put(h, computeEdgePoint(h));
		}
		
		return result;
	}
	
	/**
	 * Compute the new coordinates for a vertex (already existing in the initial mesh)
	 */
	public Point_3 computeNewVertexLocation(Vertex<Point_3> v) {
		//throw new Error("To be completed");
		
		double alpha=3./16.;
		int d=this.polyhedron3D.vertexDegree(v);
		if(d>3)
			alpha=3./(8.*d);
		
		Point_3[] neighbors=new Point_3[d+1]; // geometric coordinates of the neighbors of vertex v
		Double[] coefficients=new Double[d+1];
		
		// compute a linear combination of the neighboring vertices around v
		Halfedge<Point_3> h=v.getHalfedge(); // first halfedge incident to v
		Halfedge<Point_3> pEdge=h.getOpposite().getPrev(); // next halfedge around v, after h
		neighbors[0]=h.getOpposite().getVertex().getPoint();
		coefficients[0]=alpha;
		int i=1;
		while(pEdge!=h) { // turn around v
			neighbors[i]=pEdge.getOpposite().getVertex().getPoint();
			coefficients[i]=alpha;
			pEdge=pEdge.getOpposite().getPrev();
			i++;
		}
		
		neighbors[d]=v.getPoint(); // the last point is the location of v itself
		coefficients[d]=1-alpha*d;
		
		return Point_3.linearCombination(neighbors, coefficients);
	}

	/**
	 * Compute the new coordinates for all vertices of the initial mesh
	 */
	public Point_3[] computeNewVertexLocations() {
		//throw new Error("To be completed");
		Point_3[] result=new Point_3[this.polyhedron3D.sizeOfVertices()];
		int i=0;
		for(Vertex<Point_3> v: this.polyhedron3D.vertices) {
			result[i]=this.computeNewVertexLocation(v);
			i++;
		}
		return result;
	}

}
