package Jcg.viewer.editor2d;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import Jcg.convexhull2d.*;
import Jcg.geometry.*;
import Jcg.graph.GeometricGraph_2;
import Jcg.graph.Node;
import Jcg.polyhedron.*;
import Jcg.viewer.editor2d.Draw;

/**
 * Simple class for visualizing planar graphs and plane objects (points, segments, ...)
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique (INF562)
 * @version dec 2012
 */
public class Canvas2D extends Canvas implements MouseListener, MouseMotionListener {	
	// constants defining the action to be performed
	public static final int ADD_VERTEX=0;
	public static final int ADD_EDGE=1;
	public static final int REMOVE_VERTEX=2;
	public static final int MOVE_VERTEX=3;
	public static final int ZOOM_IN=4;
	public static final int ZOOM_OUT=5;
	public static final int MOVE_WINDOW=6;
	
	// geometric parameters defining the rendering of the scene
	public Point_2 p1,p2; // vertices of the window (x and y ranges)
	int x1,x2,y1,y2; // mouse coordinates
	public int height, width; // dimension of the window (number of pixels)
	public double epsilon; // for defining closest points
	
	public Draw draw; // class defining drawing methods
	
	public int actionToPerform; // action to be performed
	public boolean showBoundingBox; // decide whether the bounding box is shown
	public boolean showGrid; // decide whether the grid is shown

	// Geometric data (to manipulate) and algorithms
	public Polyhedron_3<Point_2> mesh=null; // polygonal planar mesh (half-edge data structure)
	public Color[] edgeColors=null; // for drawing colored edges of a mesh
	public GeometricGraph_2 graph=null; // for representing points and graphs (embedded in 2D)
	public AlgorithmList algorithms=null; // collection of available algorithms

	// Geometric objects to draw (only for rendering, not manipulation)
	private ArrayList<Segment_2> segments=null;
	private ArrayList<Color> segmentColors=null;
	private ArrayList<Point_2> points=null;
	private ArrayList<String> labels=null;
	private ArrayList<Point_2> labelPositions=null;
	
	/**
	 * Create a 2D window for drawing 2D objects
	 */
	public Canvas2D() {
		addMouseListener(this);
		addMouseMotionListener(this);
		setBackground(Color.white);
		
		Dimension d=getSize(); 
		this.height=d.height;
		this.width=d.width;
		
		this.p1=new Point_2(-1.4,-1.4);
		this.p2=new Point_2(1.4,1.4);
		this.epsilon=this.p1.distanceFrom(this.p2).doubleValue()/(100);
		
		this.graph=new GeometricGraph_2(); // initialize geometric data
		this.algorithms=new AlgorithmList(this); // initialize algorithm implementations
		this.initRenderedObjects(); // initialize geometric objects to render
		
		this.actionToPerform=ADD_VERTEX;
		this.showBoundingBox=false;
		this.showGrid=true;
	}
	
	/**
	 * Restore initial configuration (data are deleted)
	 */
	public void restore() {
		this.p1=new Point_2(-4.,-4.);
		this.p2=new Point_2(4.,4.);
		this.epsilon=this.p1.distanceFrom(this.p2).doubleValue()/(100);
		
		this.graph=new GeometricGraph_2(); // clear all data
		this.initRenderedObjects(); // clear rendered objects
		this.actionToPerform=ADD_VERTEX;
		this.showBoundingBox=false;
		this.showGrid=true;
		this.repaint();
	}
	
	/**
	 * Clear all rendered objects (geometric data are not deleted)
	 */
	public void initRenderedObjects() {
		this.points=new ArrayList<Point_2>();
		this.segments=new ArrayList<Segment_2>();
		this.segmentColors=new ArrayList<Color>();
		this.labels=new ArrayList<String>();
		this.labelPositions=new ArrayList<Point_2>();
		//this.repaint();
	}	

	/**
	 * Perform a zoom (resize the window limits)
	 */
	public void zoomIn() {
		double factor=0.1; // scaling factor (10%)
		double dx=this.p2.getX().doubleValue()-this.p1.getX().doubleValue();
		double dy=this.p2.getY().doubleValue()-this.p1.getY().doubleValue();
		
		dx=dx*factor;
		dy=dy*factor;
		
		p1.translateOf(new Vector_2(dx, dy)); // move the corners of the window
		p2.translateOf(new Vector_2(-dx, -dy));
	}

	/**
	 * Perform a reverse zoom (resize the window limits)
	 */
	public void zoomOut() {
		double factor=0.1; // scaling factor (10%)
		double dx=this.p2.getX().doubleValue()-this.p1.getX().doubleValue();
		double dy=this.p2.getY().doubleValue()-this.p1.getY().doubleValue();
		
		dx=dx*factor;
		dy=dy*factor;
		
		p1.translateOf(new Vector_2(-dx, -dy)); // move the corners of the window
		p2.translateOf(new Vector_2(+dx, +dy));
	}
	
	public void mouseClicked(MouseEvent e) {
  		int i=e.getX();
  		int j=e.getY();
  		
  		if(this.actionToPerform==ADD_VERTEX) { // add a vertex
  			this.graph.addNode(this.draw.getPoint(i, j));
  			this.algorithms.executeAlgorithm();
  		}
  		if(this.actionToPerform==REMOVE_VERTEX) { // add a vertex
  			Point_2 queryPoint=this.draw.getPoint(i, j);
  			Node<Point_2> v=findClosestVertex(this.graph, queryPoint, this.epsilon);
  			if(v!=null) {
  				this.graph.removeNode(v);
  				this.algorithms.executeAlgorithm();
  			}
  		}
  		this.repaint();
 	}
  	
  	public void mouseReleased(MouseEvent e) {
  		x2=e.getX();
  		y2=e.getY();
  		
  		if(this.actionToPerform==MOVE_VERTEX) {
  			Point_2 initialPosition=this.draw.getPoint(x1, y1);
  			Node<Point_2> v=findClosestVertex(this.graph, initialPosition, this.epsilon);
  			
  			if(v!=null) {
  				Point_2 newPosition=this.draw.getPoint(x2, y2);
  				this.graph.setPoint(v, newPosition);
  				this.algorithms.executeAlgorithm();
  			}
  		}
  		if(this.actionToPerform==ADD_EDGE) {
  			Point_2 firstPosition=this.draw.getPoint(x1, y1);
  			Point_2 secondPosition=this.draw.getPoint(x2, y2);
  			Node<Point_2> u=findClosestVertex(this.graph, firstPosition, this.epsilon);
  			Node<Point_2> v=findClosestVertex(this.graph, secondPosition, this.epsilon);
  			
  			if(u!=null && v!=null) {
  				System.out.println("Edge added");
  				this.graph.addEdge(u, v);
  			}
  		}
  		e.consume();
  		repaint();
  	}
  	
  	public void mouseEntered(MouseEvent e) {}
  	public void mouseExited(MouseEvent e) {}
  	public void mouseMoved(MouseEvent e) {}

  	public void mouseDragged(MouseEvent e) {
  	}

  	public void mousePressed(MouseEvent e) {
  		x1=e.getX(); // coordinate of the cursor, when button is pressed
  	  	y1=e.getY();
  	  	
  	  	this.repaint();
  	}

	/**
	 * Main method defining the rendering of the scene
	 */
  	public void paint(Graphics g) {
  		Dimension d=getSize(); 
  		this.height=d.height-2; 
  		this.width=d.width-2;  

  		this.draw= new Draw(g,p1,p2,height,width);
    	g.setColor(Color.black);
    	
    	if(this.showGrid==true)
    		this.draw.drawGrid();
    	
    	if(this.showBoundingBox==true) {
    		g.setColor(Color.lightGray);
    		Point_2[] box=this.graph.boundingBox();
    		this.draw.drawBoundingBox(box);
    	}
    	
    	if(this.graph!=null)
    		this.draw.drawGraph(this.graph); // draw a planar geometric graph, if defined
    	if(this.mesh!=null)
    		this.draw.drawMesh(mesh, this.edgeColors); // draw a planar mesh, if defined
    	
    	// drawing segments
    	int i=0;
    	for(Segment_2 s: this.segments) {
    		Color c=this.segmentColors.get(i);
    		this.draw.drawSegment(s, c);
    	}
    	
    	// drawing points
    	for(Point_2 p: this.points){
    		this.draw.drawCircle(p, 9, Color.red);
    	}
    	
    	// drawing labels
    	for(String label: this.labels) {
    		Point_2 p=this.labelPositions.get(i);
    		this.draw.drawLabel(p, label);
    		i++;
    	}
    	
    }

  	//--- methods for retrieving vertices and manipulating the graph -----
  	
	/**
	 * Return the list of vertices which are at distance at most epsilon from point p
	 */
  	public static Collection<Node<Point_2>> findCloseVertices(GeometricGraph_2 graph, Point_2 p, double epsilon) {
  		Collection<Node<Point_2>> result=new ArrayList<Node<Point_2>>();
  		if(graph==null || p==null)
  			return result;
  		for(Node<Point_2> v: graph.vertices) {
  			Point_2 q=graph.getPoint(v);
  			if(p.distanceFrom(q).doubleValue()<=epsilon)
  				result.add(v);
  		}
  		return result;
  	}

	/**
	 * Return the list of vertices which are at distance at most epsilon from point p
	 */
  	public static Node<Point_2> findClosestVertex(GeometricGraph_2 graph, Point_2 p, double epsilon) {
  		Node<Point_2> closestPoint=null;
 		if(graph==null || p==null)
  			return null;
   		
 		double minDistance=Double.MAX_VALUE;
 		Collection<Node<Point_2>> closeVertices=findCloseVertices(graph, p, epsilon);
  		for(Node<Point_2> v: closeVertices) {
  			Point_2 q=graph.getPoint(v);
  			if(p.distanceFrom(q).doubleValue()<=minDistance) {
  				closestPoint=v;
  				minDistance=p.distanceFrom(q).doubleValue();
  			}
  		}
  		return closestPoint;
  	}

  	//----------- methods for adding 2D objects to the drawing -----------
  	
  	public void addSegment(Point_2 p, Point_2 q) {
  		if(p==null || q==null) {
  			System.out.println("error null point");
  			return;
  		}
  		this.segments.add(new Segment_2(p, q));
  		this.segmentColors.add(Color.black);
  	}

  	public void addSegment(Point_2 p, Point_2 q, Color c) {
  		if(p==null || q==null) {
  			throw new Error("error: addSegment, null point");
  		}
  		if(c==null) {
  			throw new Error("error: addSegment, null color");
  		}
  		this.segments.add(new Segment_2());
  		this.segmentColors.add(c);
  	}
  	
  	public void addSegment(Segment_2 s) {
  		if(s==null) {
  			System.out.println("error null segment");
  			return;
  		}
  		this.segments.add(s);
  		this.segmentColors.add(Color.black);
  	}

  	public void addSegment(Segment_2 s, Color c) {
  		if(s==null)
  			return;
  		this.segments.add(s);
  		this.segmentColors.add(c);
  	}
  	
  	public void addPoint(Point_2 p) {
  		if(p==null)
  			return;
  		this.points.add(p);
  	}
  	
  	public void addPoints(Collection<Point_2> points) {
  		if(points==null)
  			return;
  		this.points.addAll(points);
  	}
  	
  	public void addSegments(Collection<Segment_2> segments) {
  		if(points==null)
  			return;
  		for(Segment_2 s: segments) {
  			this.addSegment(s);
  		}
  	}
  	
  	public void addLabel(Point_2 p, String label) {
  		if(p==null || label==null)
  			return;
  		this.labels.add(label);
  		this.labelPositions.add(p);
  	}
  	
  	public void addPolyhedronEdges(Polyhedron_3<Point_2> mesh, Color[] edgeColors) {
  		if(mesh==null) return;
  		
  		if(edgeColors==null) {
  			for(Halfedge<Point_2> e: mesh.halfedges) {
  				Point_2 p=e.getVertex().getPoint();
  				Point_2 q=e.getOpposite().getVertex().getPoint();
  				if(p==null || q==null)
  					throw new Error("error: null point");
  				this.addSegment(p, q);
  			}
  		}
  		else {
  			int i=0;
  			for(Halfedge<Point_2> e: mesh.halfedges) {
  				Point_2 p=e.getVertex().getPoint();
  				Point_2 q=e.getOpposite().getVertex().getPoint();
  				if(p==null || q==null)
  					throw new Error("error: null point");
  				this.addSegment(p, q, edgeColors[i]);
  				i++;
  			}
  		}
  	}

  	public void addPolyhedronVertices(Polyhedron_3<Point_2> mesh) {
  		if(mesh==null) return;
  		
  		for(Vertex<Point_2> v: mesh.vertices)
  			this.addPoint(v.getPoint());
  		
  	}

  	public void addPolyhedronVertexLabels(Polyhedron_3<Point_2> mesh, String[] labels) {
  		if(mesh==null || labels==null) return;
  		
  		int i=0;
  		for(Vertex<Point_2> v: mesh.vertices) {
  			if(labels[i]!=null)
  				this.addLabel(v.getPoint(), labels[i]);
  			i++;
  		}
  		
  	}

}
 